# ruff: noqa: E501 Line too long
import os
import sys
import json
from argparse import ArgumentParser, BooleanOptionalAction, _SubParsersAction
from pathlib import Path
from typing import Generator
import argcomplete
from . import __version__ as moulti_version
from .client import current_instance, moulti_socket_path, send_to_moulti, send_to_moulti_and_handle_reply, pipeline
from .environ import env, pint, float_str
from .widgets.cli import add_cli_arguments
from .manpage import manpage_parse, manpage_run

def init(args: dict) -> None:
	"""Start a new Moulti instance."""

	from .app import main as init_moulti # pylint: disable=import-outside-toplevel
	init_moulti(**args)

def moulti_run_should_suffix_instance_name(args: dict) -> bool:
	"""
	By default, `moulti run` suffixes the instance name with the process id.
	Since the socket path is derived from the instance name, this helps prevent clashes and thus
	"cannot listen" errors.
	"""
	if env('MOULTI_SOCKET_PATH'):
		# If a socket path was explicitly set, then the instance name cannot affect
		# its computation. Therefore, there is no need to suffix the instance name.
		return False
	if args['no_suffix']:
		return False
	if env('MOULTI_RUN_NO_SUFFIX') is not None:
		return False
	return True

def run(args: dict) -> None:
	"""Start a new Moulti instance and run the given command."""
	if moulti_run_should_suffix_instance_name(args):
		os.environ['MOULTI_INSTANCE'] = f'{current_instance()}-{os.getpid()}'

	# Handle --print-env:
	if args['print_env']:
		from .app import run_environment # pylint: disable=import-outside-toplevel
		environment_variables = run_environment(args['command'], moulti_socket_path(), False)
		for name, value in environment_variables.items():
			print(f'{name}={value}')
		sys.exit(0)

	from .app import main as init_moulti # pylint: disable=import-outside-toplevel
	init_moulti(args['command'])

def wait(verbose: bool = False, delay: int = 500, max_attempts: int = 0) -> None:
	"""Wait until the Moulti instance is available.
	Args:
		verbose: if True, output the reason why each connection attempt failed
		delay: number of milliseconds between two connection attempts
		max_attempts: maximum number of attempts before giving up; 0 means "never give up"
	"""
	import time # pylint: disable=import-outside-toplevel
	connected = False
	attempts = 0
	while not connected:
		try:
			attempts += 1
			send_to_moulti({'command': 'ping'})
			connected = True
			break
		except Exception as exc:
			if verbose:
				print(f'Connection #{attempts} to {moulti_socket_path()}: {exc}')
			if max_attempts > 0 and attempts == max_attempts:
				print('Giving up.')
				break
			time.sleep(delay / 1000.0)
	sys.exit(0 if connected else 1)

def saved_files(directory: Path) -> Generator:
	"""
	Iterate over a directory supposedly generated by Moulti's "Save" feature.
	Yield pipeline()-compatible triplets.
	"""
	json_ext = '.properties.json'
	for json_file in sorted(directory.glob('*' + json_ext)):
		if not json_file.is_file():
			continue
		with json_file.open(encoding='utf-8', errors='surrogateescape') as json_file_desc:
			data = json.load(json_file_desc)
			log_file = json_file.with_name(json_file.name[:-len(json_ext)] + '.contents.log')
			fileno = os.open(str(log_file), os.O_RDONLY) if log_file.is_file() else None
			yield data.get('id'), data, fileno

def load(args: dict, read_size: int = 1024**2) -> None:
	"""
	Load the contents of a directory supposedly generated by Moulti's "Save" feature and push it to a Moulti instance.
	"""
	errors = pipeline(saved_files(args['saved_directory']), read_size)
	sys.exit(errors)

def diff_parse(args: dict) -> None:
	from .diff import diff_parse as real_diff_parse # pylint: disable=import-outside-toplevel
	real_diff_parse(args)

def diff_run(args: dict) -> None:
	from .diff import diff_run as real_diff_run # pylint: disable=import-outside-toplevel
	real_diff_run(args)

def add_main_commands(subparsers: _SubParsersAction) -> None:
	# moulti init
	init_parser = subparsers.add_parser('init', help='start a new Moulti instance')
	init_parser.set_defaults(func=init)

	# moulti run
	run_parser = subparsers.add_parser('run', help='start a new Moulti instance and run a command')
	run_parser.set_defaults(func=run)
	run_parser.add_argument('--print-env', action='store_true', default=False, help='print environment variables set by Moulti and exit')
	run_parser.add_argument('--no-suffix', '-n', action='store_true', default=False, help='do not suffix the instance name with the process id')
	run_parser.add_argument('command', type=str, nargs='+', help='command to run along with its arguments')

	# moulti wait
	wait_parser = subparsers.add_parser('wait', help='wait until the Moulti instance is available')
	wait_parser.set_defaults(func=wait)
	wait_parser.add_argument('--verbose', '-v', action='store_true', help='if True, output the reason why each connection attempt failed')
	wait_parser.add_argument('--delay', '-d', type=pint, default=500, help='number of milliseconds between two connection attempts')
	wait_parser.add_argument('--max-attempts', '-m', type=pint, default=0, help='maximum number of attempts before giving up; 0 means "never give up"')

	# moulti set
	set_parser = subparsers.add_parser('set', help='set Moulti options')
	set_parser.set_defaults(func=send_to_moulti_and_handle_reply, command='set')
	set_parser.add_argument('--title', '-t', default=None, type=str, help='title displayed at the top of the screen')
	set_parser.add_argument('--step-position', default=None, type=str, choices=('top', 'bottom'), help='whether to display steps at the top (default) or bottom of the screen')
	set_parser.add_argument('--step-direction', default=None, type=str, choices=('up', 'down'), help='whether to lay out steps up or down (default)')
	set_parser.add_argument('--progress-bar', default=None, action=BooleanOptionalAction, help='whether to display the progress bar')
	set_parser.add_argument('--progress-target', '-pt', default=None, type=float, help='total number of steps associated with the progress bar')
	set_parser.add_argument('--progress', '-p', default=None, type=float_str, help='progress so far, in number of steps; accept absolute or relative values, e.g. 50, +1 or -5')

	# moulti load
	load_parser = subparsers.add_parser('load', help='load a saved directory into Moulti')
	load_parser.set_defaults(func=load, command='load')
	arg = load_parser.add_argument('saved_directory', type=Path, help='path to a directory generated by Moulti\'s "Save" feature')
	arg.completer = argcomplete.completers.DirectoriesCompleter()

	# moulti diff parse/run
	diff_parser = subparsers.add_parser('diff', help='load unified diff data into Moulti')
	diff_subparsers = diff_parser.add_subparsers(required=True)
	diff_parse_parser = diff_subparsers.add_parser('parse', help='load unified diff data into Moulti from a file')
	diff_parse_parser.set_defaults(func=diff_parse)
	arg = diff_parse_parser.add_argument('diff_filepath', type=Path, help='path to a unified diff file')
	arg.completer = argcomplete.completers.FilesCompleter()
	diff_run_parser = diff_subparsers.add_parser('run', help='load unified diff data into Moulti from a command')
	diff_run_parser.set_defaults(func=diff_run)
	diff_run_parser.add_argument('command', type=str, nargs='+', help='command to run along with its arguments')

	# moulti manpage parse/run
	manpage_parser = subparsers.add_parser('manpage', help='load man page data into Moulti')
	manpage_subparsers = manpage_parser.add_subparsers(required=True)
	manpage_parse_parser = manpage_subparsers.add_parser('parse', help='load man page data into Moulti from a file')
	manpage_parse_parser.set_defaults(func=manpage_parse)
	arg = manpage_parse_parser.add_argument('manpage_filepath', type=Path, help='path to a file holding man output')
	arg.completer = argcomplete.completers.FilesCompleter()
	manpage_run_parser = manpage_subparsers.add_parser('run', help='load man page data into Moulti from a command')
	manpage_run_parser.set_defaults(func=manpage_run)
	manpage_run_parser.add_argument('command', type=str, nargs='+', help='command to run along with its arguments')

	# moulti scroll
	scroll_parser = subparsers.add_parser('scroll', help='scroll to make a specific step visible')
	scroll_parser.set_defaults(func=send_to_moulti_and_handle_reply, command='scroll')
	scroll_parser.add_argument('id', type=str, help='step identifier')
	scroll_parser.add_argument('offset', type=int, nargs='?', default=None, help='-1: last line, 0: first line, 1: second line, etc.')

def build_arg_parser() -> ArgumentParser:
	arg_parser = ArgumentParser(prog='moulti', description='step-by-step logs')
	arg_parser.add_argument('--version', action='version', version=moulti_version)
	subparsers = arg_parser.add_subparsers(required=True)
	# moulti init, moulti wait:
	add_main_commands(subparsers)
	# moulti <widget>:
	add_cli_arguments(subparsers)
	return arg_parser

def first_non_option_argument(args: list[str], start: int = 0) -> int|None:
	"""
	Return the index of the first non-option argument, i.e. the first argument that does not start with a "-" (dash)
	character.
	"""
	for index, arg in enumerate(args[start:], start):
		if not arg.startswith('-'):
			return index
	return None

def inject_double_dash_before_command(args: list[str], start: int = 0) -> None:
	"""
	Inject "--" right before the first non-option argument.
	"""
	index = first_non_option_argument(args, start)
	if index is not None and index > 0 and args[index - 1] != '--':
		args.insert(index, '--')

def adjust_cli_args(args: list[str]) -> None:
	"""
	Command-line arguments are parsed using argparse. Consequently, "run" subcommands often require users to add "--" to
	their command line, e.g. "moulti run -- ls -al" instead of "moulti run ls -al".
	Detect such cases and inject "--" if it is missing.
	"""
	if len(args) < 2:
		return
	if args[1] == 'run':
		inject_double_dash_before_command(args, 2)
	elif args[1] in ('manpage', 'diff') and args[2] == 'run':
		inject_double_dash_before_command(args, 3)

def main() -> None:
	try:
		arg_parser = build_arg_parser()
		argcomplete.autocomplete(arg_parser, always_complete_options='long')
		adjust_cli_args(sys.argv)
		args = vars(arg_parser.parse_args())
		func = args.pop('func')
		# Subtlety: func and args are not always used the same way:
		if func == wait: # pylint: disable=comparison-with-callable
			wait(**args)
		else:
			func(args)
	except KeyboardInterrupt:
		print('')
		sys.exit(1)

if __name__ == '__main__':
	main()



"""
Provide everything to get settings from environment variables.
"""
from os import environ
from os.path import isfile
from typing import Any, Type, TypeAlias, Callable, Iterable
from argparse import ArgumentTypeError

AcceptableType: TypeAlias = Type|Callable[[str], Any]
AcceptableTypes: TypeAlias = Iterable[AcceptableType]
OptionalAcceptableTypes: TypeAlias = AcceptableTypes|None


def pint(value: str) -> int:
	integer_value = int(value)
	if integer_value < 0:
		raise ArgumentTypeError('expected a positive integer')
	return integer_value


def spint(value: str) -> int:
	integer_value = int(value)
	if integer_value <= 0:
		raise ArgumentTypeError('expected a strictly positive integer')
	return integer_value


def float_str(value: str) -> str:
	_ = float(value)
	return value


def bool_or_int(value: str) -> bool|int:
	if value.lower() == 'false':
		return False
	if value.lower() == 'true':
		return True
	return int(value)


def existing_filepath(value: str) -> str:
	if not isfile(value):
		raise ArgumentTypeError('expected an existing filepath')
	return value


def try_types(value: Any, types: AcceptableTypes) -> tuple[Any, bool]:
	for acceptable_type in types:
		try:
			return acceptable_type(value), True
		except Exception:
			pass
	return value, False


def env(
	name: str,
	default: Any = None,
	types: OptionalAcceptableTypes = None,
) -> Any:
	"""
	Return environment variable name and ensure it fits one of the provided types.
	Otherwise, return the default value.
	"""
	if (value := environ.get(name)) is not None:
		if types is None:
			return value
		value, ok = try_types(value, types)
		if ok:
			return value
	return default


def enva(
	name: str,
	default: Any = None,
	types: OptionalAcceptableTypes = None,
	sep: str = ',',
	strict: bool = True,
) -> Any:
	"""
	Get an array from the environment.
	Specifically, get environment variable name, split it using separator sep and ensure each resulting item fits one of
	the provided types.
	If strict is False and an item does not fit any type, it is discarded from the returned array.
	If strict is True and an item does not fit any type, this function returns the default value.
	"""
	if (value := environ.get(name)) is None:
		return default
	if not value:
		return []
	values = value.split(sep)
	if types is None:
		return values

	final_list = []
	for item in values:
		value, ok = try_types(item, types)
		if ok:
			final_list.append(value)
		elif strict:
			return default
	return final_list


def envd(
	name: str,
	default: Any = None,
	types: OptionalAcceptableTypes = None,
	sepa: str = ',',
	sepkv: str = '=',
	strict: bool = True,
) -> Any:
	"""
	Get a dict from the environment.
	Specifically, get environment variable name, split it using array separator sepa, split each resulting item using
	key-value separator sepkv then ensure each resulting value fits one of the provided types.
	If strict is False and a value does not fit any type, the key-value pair is discarded from the returned array.
	If strict is True and a value does not fit any type, this function returns the default value.
	"""
	key_values = enva(name, default=None, types=None, sep=sepa)
	if key_values is None:
		# enva() returned the default value, and so does this function:
		return default
	final_dict = {}
	for key_value in key_values:
		if sepkv in key_value:
			key, value = key_value.split(sepkv, maxsplit=1)
		else:
			return default
		if types is None:
			final_dict[key] = value
		else:
			value, ok = try_types(value, types)
			if ok:
				final_dict[key] = value
			elif strict:
				return default
	return final_dict



import re
import sys
import subprocess
from io import StringIO
from os import environ, getpid
from typing import Any, Generator
from rich.text import Text
from .pipeline import pipeline
from .environ import env

ENCODING = env('MOULTI_MANPAGE_ENCODING') or 'utf-8'

def overtype_to_ansi(content: str) -> str:
	"""
	Under *BSD and macOS, man pages are formatted using overtypes, like an
	actual teleprinter would do.
	Rich/Textual handle ANSI escape sequences but not overtypes.
	"""
	# character, backspace, character = bold character
	content = re.sub(r'(.)\x08\1', '\x1b[1m\\1\x1b[0m', content)
	# underscore, backspace, character = underlined character
	content = re.sub(r'_\x08(.)', '\x1b[4m\\1\x1b[24m', content)
	return content

def dissect_manpage(filedesc: Any) -> dict:
	title = ''
	while line := filedesc.readline():
		line = line.strip('\n')
		if line:
			title = overtype_to_ansi(line)
			break
	all_lines = [overtype_to_ansi(line) for line in filedesc.readlines()]
	footer = all_lines.pop().strip('\n')

	sections: list[dict] = []
	current_section = {'title': '', 'text': ''}

	def add_section(section: dict[str, str]) -> None:
		section['text'] = section['text'].strip('\n')
		if section['title'] or section['text']:
			sections.append(section)

	for line in all_lines:
		if line.startswith(' ') or line == '\n': # regular line
			current_section['text'] += line
		else: # section title
			add_section(current_section)
			current_section = {'title': line.strip('\n'), 'text': ''}
	add_section(current_section) # Handle the last section

	return {
		'title': title,
		'sections': sections,
		'footer': footer,
	}

def line_indent_size(line: str) -> int:
	size = 0
	for char in line:
		if char != ' ':
			break
		size += 1
	return size

def section_indent_size(text: str) -> int|None:
	size = None
	for line in text.split('\n'):
		if not line:
			continue
		lis = line_indent_size(line)
		if size is None or lis < size:
			size = lis
	return size

def unindent(text: str, max_size: int = -1) -> str:
	size = section_indent_size(text)
	if size is None:
		return text
	if max_size >= 0:
		size = min(size, max_size)
	result = []
	for line in text.split('\n'):
		if line:
			line = line[size:]
		result.append(line)
	return '\n'.join(result)

def ansi_to_markup(text: str) -> str:
	return Text.from_ansi(text).markup

def commands(title: str, manpage: dict) -> Generator:
	"""
	Yield pipeline()-compatible triplets.
	"""
	if env('MOULTI_MANPAGE_NO_TITLE') is None:
		yield None, {'command': 'set', 'title': title}, None

	print_steps = bool(env('MOULTI_MANPAGE_VERBOSE'))
	pid = getpid()
	counter = {'step': 1}
	def step(cmd: str, title: str, **kwargs: Any) -> tuple[str, dict, None]:
		step_id = f'manpage_{pid}_{counter["step"]}'
		counter['step'] += 1
		step = {'command': cmd, 'action': 'add', 'id': step_id, 'title': title, 'classes': 'manpage', **kwargs}
		if print_steps:
			print(f'{cmd} {step_id}')
		return step_id, step, None

	yield step('divider', ansi_to_markup(manpage['title']), classes='manpage title')
	max_unindent = 3
	for section in manpage['sections']:
		text = unindent(section['text'], max_unindent)
		yield step('step', ansi_to_markup(section['title']), text=text, collapsed=False, max_height=0)
	yield step('divider', ansi_to_markup(manpage['footer']), classes='manpage footer')

def handle_manpage_from_file_descriptor(title: str, filedesc: Any) -> None:
	pipeline(commands(title, dissect_manpage(filedesc)))

def handle_manpage_from_string(title: str, data: str) -> None:
	handle_manpage_from_file_descriptor(title, StringIO(data))

def manpage_parse(args: dict) -> None:
	filepath = args['manpage_filepath']
	with open(filepath, 'r', encoding=ENCODING, errors='surrogateescape') as filedesc:
		title = f'{filepath}'
		handle_manpage_from_file_descriptor(title, filedesc)

def manpage_run(args: dict) -> None:
	command = args['command']
	environ['MAN_KEEP_FORMATTING'] = 'y'
	res = subprocess.run(command, check=False, stdout=subprocess.PIPE, encoding=ENCODING, errors='surrogateescape')
	if res.stdout:
		title = ' '.join(command)
		handle_manpage_from_string(title, res.stdout)
	if res.returncode and not res.stdout:
		sys.stderr.write(f'manpage_run: {command} produced no output and exited with status {res.returncode}\n')
		sys.exit(res.returncode)
	sys.exit(0)



import re
from collections import deque
from dataclasses import dataclass
from typing import Any
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.message import Message
from textual.reactive import reactive
from textual.validation import Function
from textual.widgets import Input, Label, Static
from moulti.search import TextSearch

def regex_is_valid(regex: str) -> bool:
	try:
		re.compile(regex)
		return True
	except re.error:
		return False

REGEX_VALIDATOR = Function(regex_is_valid)
SEARCH_HISTORY_SIZE = 500 # The search history is not persisted so this should be plenty.

class SearchInput(Input):
	class CancelSearch(Message):
		pass

	@dataclass
	class HistoryMove(Message):
		move: int

	BINDINGS = [
		Binding('escape', 'exit', 'Cancel', priority=True),
		Binding('up', 'history(-1)', 'Previous', priority=True),
		Binding('down', 'history(+1)', 'Next', priority=True),
	]

	def action_exit(self) -> None:
		self.post_message(self.CancelSearch())

	def action_delete_left(self) -> None:
		"""
		Hitting backspace when the input is empty cancels the search.
		"""
		if self.value:
			super().action_delete_left()
		else:
			self.action_exit()

	async def action_submit(self) -> None:
		"""
		Hitting enter when the input is empty cancels the search.
		"""
		if self.value:
			await super().action_submit()
		else:
			self.action_exit()

	def action_history(self, move: int) -> None:
		self.post_message(self.HistoryMove(move))

	DEFAULT_CSS = """
	SearchInput {
		background: $boost;
		color: $text;
		padding: 0;
		border: none;
		width: 1fr;
		height: 1;
	}
	SearchInput:focus {
		border: none;
	}
	SearchInput.-invalid {
		border: none;
		color: $error;
	}
	SearchInput.-invalid:focus {
		border: none;
		color: $error;
	}
	"""


class SearchInputWidget(Static):
	class NewSearch(Message):
		def __init__(self, search: TextSearch):
			super().__init__()
			self.search = search

	LABELS = {
		(True, True): 'REGEX',
		(True, False): 'Regex',
		(False, True): 'PLAIN',
		(False, False): 'Plain',
	}
	TOOLTIPS = {
		(True, True): 'Case-insensitive Python regular expression pattern',
		(True, False): 'Case-sensitive Python regular expression pattern',
		(False, True): 'Case-insensitive plain pattern',
		(False, False): 'Case-sensitive plain pattern',
	}

	regex = reactive(True, repaint=True, bindings=True)
	case_insensitive = reactive(False, repaint=True, bindings=True)
	next_result = reactive(True, repaint=True)

	BINDINGS = [
		Binding('ctrl+r', 'toggle_regex(True)', 'Regex pattern'),
		Binding('ctrl+r', 'toggle_regex(False)', 'Plain pattern'),
		Binding('ctrl+t', 'toggle_case_insensitive(True)', 'Case-insensitive'),
		Binding('ctrl+t', 'toggle_case_insensitive(False)', 'Case-sensitive'),
	]

	def action_toggle_regex(self, new: bool) -> None:
		self.regex = new

	def action_toggle_case_insensitive(self, new: bool) -> None:
		self.case_insensitive = new

	def check_action(self, action: str, parameters: tuple[object, ...]) -> bool:
		enable = parameters[0] is True
		if action.endswith('_regex'):
			return enable ^ bool(self.regex)
		if action.endswith('_case_insensitive'):
			return enable ^ bool(self.case_insensitive)
		return True

	def update_regex_label(self) -> None:
		configuration = (self.regex, self.case_insensitive)
		self.regex_label.update(self.LABELS[configuration])
		self.regex_label.tooltip = self.TOOLTIPS[configuration]

	def watch_regex(self, _old: bool, new: bool) -> None:
		self.update_regex_label()
		self.regex_label.set_class(new, 'regex')
		self.input.validators = [REGEX_VALIDATOR] if new else []
		self.input.validate(self.input.value)

	def watch_case_insensitive(self, _old: bool, _new: bool) -> None:
		self.update_regex_label()

	def watch_next_result(self, _old: bool, new: bool) -> None:
		self.next_label.set_class(new, 'next').update('/' if new else '?')

	def __init__(self, *args: Any, **kwargs: Any) -> None:
		super().__init__(*args, **kwargs)
		self.input = SearchInput(id='search_input')
		self.next_label = Label('?', id='search_next_label')
		self.regex_label = Label('?????', id='search_regex_label')
		self.history: deque[TextSearch] = deque(maxlen=SEARCH_HISTORY_SIZE)
		self.editable_history: deque[TextSearch] = deque(maxlen=SEARCH_HISTORY_SIZE + 1)
		self.history_index = -1

	def on_mount(self) -> None:
		self.input.valid_empty = True
		self.input.validate_on = {'changed', 'submitted'}
		self.next_label.tooltip = '/: search forward\n?: search backward'

	def compose(self) -> ComposeResult:
		with Horizontal():
			yield self.regex_label
			yield self.next_label
			yield self.input

	def to_search(self) -> TextSearch:
		return TextSearch(self.input.value, self.regex, self.case_insensitive, self.next_result)

	def setup_history(self) -> None:
		# Copy the history into the editable history:
		self.editable_history.clear()
		self.editable_history.extend(entry.copy() for entry in self.history)
		# Append a new empty entry:
		self.editable_history.append(TextSearch('', self.regex, self.case_insensitive, self.next_result))
		# This new entry is at index -1:
		self.history_index = -1

	def action_pop(self, next_result: bool = True) -> None:
		self.next_result = next_result
		self.setup_history()
		self.styles.display = 'block'
		self.input.focus()

	def set_search(self, search: TextSearch) -> None:
		self.regex = search.regex
		self.case_insensitive = search.case_insensitive
		# NOT setting self.next_result.
		self.input.value = search.pattern
		self.input.validate(self.input.value)

	def set_search_by_index(self, index: int = -1) -> None:
		if index >= 0:
			return
		try:
			self.set_search(self.editable_history[index])
			self.history_index = index
		except IndexError:
			pass

	def action_exit(self) -> None:
		self.input.clear()
		self.styles.display = 'none'

	async def on_search_input_cancel_search(self, _event: SearchInput.CancelSearch) -> None:
		self.action_exit()

	def on_search_input_history_move(self, event: SearchInput.HistoryMove) -> None:
		self.editable_history[self.history_index] = self.to_search()
		self.set_search_by_index(self.history_index + event.move)

	async def on_input_submitted(self, event: Input.Submitted) -> None:
		if not event.value:
			self.action_exit()
		else:
			if event.validation_result is None or event.validation_result.is_valid:
				search = self.to_search()
				self.action_exit()
				self.post_message(self.NewSearch(search))
				self.history.append(search)

	DEFAULT_CSS = """
		#search_regex_label {
			background: $secondary-darken-2;
			color: $text-muted;
		}
		#search_regex_label.regex {
			color: $accent;
		}
		SearchInputWidget {
			display: none;
			height: 1;
		}
		SearchInputWidget:focus, SearchInputWidget:focus-within {
			display: block;
			height: 1;
		}
	"""



import os
from selectors import BaseSelector
from typing import Any
from .environ import env, pint

def abridge_string(value: str, threshold: int = 100, sep: str = '...') -> str:
	"""
	Abridge a string longer than threshold; the result reflects the start middle and end, separated with the given
	separator.
	"""
	if len(value) <= threshold:
		return value
	sep_len = 2 * len(sep) # the abridged string features 3 parts and thus 2 separators
	chars_len = threshold - sep_len
	part_len = chars_len // 3
	delta = chars_len % 3
	middle = int((len(value) / 2) - (part_len / 2))
	return value[:part_len+delta] + sep + value[middle:middle+part_len] + sep + value[-part_len:]

def abridge_dict(message: dict[str, Any], threshold: int = 100, sep: str = '...') -> dict[str, Any]:
	"""
	Abridge the string values of a given dict using abridge_string().
	This function only processes first-level string values and does not work recursively.
	"""
	long_keys = []
	for key, value in message.items():
		if isinstance(value, str) and len(value) > threshold:
			long_keys.append(key)
	if not long_keys:
		return message
	abridged_message = {}
	for key, value in message.items():
		abridged_message[key] = abridge_string(value, threshold, sep) if key in long_keys else value
	return abridged_message

def call_all(calls: list[Any]) -> None:
	for call in calls:
		call[0](*call[1:])

def clean_selector(selector: BaseSelector|None, close_fds: bool = False, close: bool = True) -> None:
	"""
	Clean a Selector object, as offered by the selectors module, by systematically unregistering and optionally
	closing all registered file descriptors then closing the Selector object itself.
	"""
	if not selector:
		return
	for selectorkey in list(selector.get_map().values()):
		selector.unregister(selectorkey.fileobj)
		if close_fds:
			if hasattr(selectorkey.fileobj, 'close'):
				selectorkey.fileobj.close()
			elif isinstance(selectorkey.fileobj, int):
				os.close(selectorkey.fileobj)
	if close:
		selector.close()

DEFAULT_TAB_SIZE = 8

def get_tab_size() -> int:
	return env('MOULTI_TAB_SIZE', DEFAULT_TAB_SIZE, types=(pint,))

TAB_SIZE = get_tab_size()
TAB_SPACES_STR = ' '*TAB_SIZE
TAB_SPACES_BYTES = b' '*TAB_SIZE

ANSI_ESCAPE_SEQUENCE_STR = '\x1b'
ANSI_ESCAPE_SEQUENCE_BYTES = b'\x1b'

ANSI_RESET_SEQUENCES_STR = ('\x1b[0m', '\x1b[m')
ANSI_RESET_SEQUENCES_BYTES = (b'\x1b[0m', b'\x1b[m')



from .common import moulti_test
assert moulti_test

def do_test(moulti_test, monkeypatch, value, extra_args = None):
	monkeypatch.setenv('MOULTI_RUN_OUTPUT', value)
	command = ['run-output.bash']
	if extra_args is not None:
		command += extra_args
	assert moulti_test(command=command)

def test_discard_output(moulti_test, monkeypatch):
	do_test(moulti_test, monkeypatch, 'discard')

def test_harvest_output(moulti_test, monkeypatch):
	do_test(moulti_test, monkeypatch, 'harvest')

def test_custom_step(moulti_test, monkeypatch):
	do_test(moulti_test, monkeypatch, 'harvest', ['--create-step'])



from shutil import which
from .common import moulti_test
assert moulti_test

COMMAND = ['diff.bash']

def test_diff_no_delta(moulti_test, monkeypatch):
	monkeypatch.setenv('MOULTI_DIFF_NO_DELTA', 'yes')
	assert moulti_test(command=COMMAND)

if which('delta'):
	def test_diff_with_delta(moulti_test):
		assert moulti_test(command=COMMAND)



from time import sleep
from .common import Beacon, moulti_test, wait_script_start
assert moulti_test

def moulti_test_child_beacon(moulti_test, monkeypatch, expectation):
	beacon = Beacon(monkeypatch)

	def check_child_beacon(moulti_app):
		# Moulti must have exited:
		assert moulti_app.return_code == 0
		sleep(1)
		assert beacon.lit() == expectation

	assert moulti_test(
		command=['sleep.bash', '10'],
		run_before=beacon.wait,
		press='q',
		run_after=check_child_beacon
	)

def test_quit_default_ask(moulti_test, monkeypatch):
	monkeypatch.setenv('MOULTI_QUIT_POLICY', 'default=ask')
	assert moulti_test(press='q')

def test_quit_default_quit(moulti_test, monkeypatch):
	monkeypatch.setenv('MOULTI_QUIT_POLICY', 'default=quit')
	def run_after(moulti_app):
		assert moulti_app.return_code == 0
	assert moulti_test(press=('q'), run_after=run_after)

def test_quit_running_ask(moulti_test, monkeypatch):
	monkeypatch.setenv('MOULTI_QUIT_POLICY', 'running=ask')
	assert moulti_test(command=['sleep.bash', '1'], run_before=wait_script_start, press='q')

def test_quit_running_terminate(moulti_test, monkeypatch):
	monkeypatch.setenv('MOULTI_QUIT_POLICY', 'running=terminate')
	moulti_test_child_beacon(moulti_test, monkeypatch, False)

def test_quit_running_leave(moulti_test, monkeypatch):
	monkeypatch.setenv('MOULTI_QUIT_POLICY', 'running=leave')
	moulti_test_child_beacon(moulti_test, monkeypatch, True)



import os
import sys
from threading import Thread
from typing import Any, Generator
from queue import Queue
from .protocol import moulti_connect, send_json_message, recv_json_message, MoultiConnectionClosedException, Message

def pipeline(messages: Generator, read_size: int = 1024**2) -> int:
	"""
	Pipeline multiple messages to a Moulti instance.
	messages should be a generator that yields triplets:
	- step id (str; None for instance properties)
	- Moulti message to send (dict[str, Any])
	- file descriptor number (int) if it is necessary to pass data, None otherwise
	read_size is used for all "pass" operations.
	Note: so far, this is the only function in Moulti that pipelines messages and replies. Therefore, it is also the
	only function that actually leverages message ids.
	"""
	sender_thread: Queue = Queue()
	receiver_thread = sender_thread
	filenos = []
	sent_messages: dict[str, bool] = {}
	stats = {'errors': 0}
	unexpected_replies = []

	def send_message(*args: Any) -> None:
		msg_id = send_json_message(*args)
		sent_messages[msg_id] = False
		receiver_thread.put_nowait(msg_id)

	def process_reply(reply: Message, store: bool = False) -> bool:
		if 'msgid' in reply:
			if reply['msgid'] in sent_messages:
				sent_messages[reply['msgid']] = True
				if reply.get('done') is not True:
					stats['errors'] += 1
					sys.stderr.write(f'Error: {reply.get("error")}\n')
				return True
		if store:
			unexpected_replies.append(reply)
		return False

	def recheck_unexpected_replies() -> None:
		if unexpected_replies:
			for index in range(len(unexpected_replies) - 1, -1, -1):
				if process_reply(unexpected_replies[index], store=False):
					del unexpected_replies[index]

	with moulti_connect() as moulti_socket:
		# Spawn a thread that receives as many replies as sent messages:
		def recv_messages() -> None:
			try:
				while sender_thread.get() is not None:
					recheck_unexpected_replies()
					reply, _ = recv_json_message(moulti_socket, 0)
					process_reply(reply, store=True)
			except MoultiConnectionClosedException:
				pass
		recv_msg_thread = Thread(target=recv_messages)
		recv_msg_thread.start()

		for step_id, data, fileno in messages:
			# Send all messages using protocol pipelining:
			try:
				# Properties, typically a command that creates and configures a step:
				send_message(moulti_socket, data)
				# Contents:
				if fileno is not None:
					filenos.append(fileno)
					pass_msg = {'command': 'pass', 'id': step_id, 'read_size': read_size}
					send_message(moulti_socket, pass_msg, [fileno])
			except Exception as exc:
				stats['errors'] += 1
				sys.stderr.write(f'Error with {step_id}/{data}/{fileno}: {exc}\n')
		receiver_thread.put_nowait(None)
		recv_msg_thread.join()
		recheck_unexpected_replies()
		for fileno in filenos:
			os.close(fileno)
	return stats['errors']



from textual.app import ComposeResult
from textual.widgets import Footer as TextualFooter

class Footer(TextualFooter):
	"""
	Override the (black) Footer class so it looks like the (blue) ClassicFooter.
	"""

	def compose(self) -> ComposeResult:
		"""
		Footer <  0.63.0 is rendered, not composed, so this method has no practical effect.
		Footer >= 0.63.0 is composed, and this method prepends the description
		with an extra space to closely imitate the look of the previous Footer.
		"""
		for footer_key in super().compose():
			if hasattr(footer_key, 'description'):
				footer_key.description = ' ' + footer_key.description
			yield footer_key

	FOOTER_KEY_CSS = """
		background: $primary;
		color: $footer-key-foreground;
		.footer-key--description {
			color: $footer-key-foreground;
		}
		.footer-key--key {
			background: $primary-darken-2;
			color: $footer-key-foreground;
		}
		&:hover {
			background: $primary-darken-1;
			.footer-key--key {
				background: $accent;
				color: $footer-key-foreground;
			}
		}
		&.-disabled {
			text-style: dim;
			background: $primary;
			&:hover {
				.footer-key--key {
					background: $primary-darken-2;
				}
			}
		}"""

	DEFAULT_CSS = """
Footer {
	color: $footer-key-foreground;
    background: $primary;
	/* FooterKey is specific to the new Footer and thus does not affect the previous one. */
	FooterKey {%s
	}
	FooterKey:light {%s
	}
}
""" % (FOOTER_KEY_CSS, FOOTER_KEY_CSS)



import os
from selectors import DefaultSelector, EVENT_READ, EVENT_WRITE
from socket import socket as Socket
from queue import Queue, Empty
from typing import Callable
from .helpers import abridge_dict, clean_selector
from .protocol import MoultiTLVReader, MoultiTLVWriter
from .protocol import MoultiProtocolException, MoultiConnectionClosedException
from .protocol import clean_socket, current_instance, data_to_message, moulti_socket_path, FDs, from_printable
from .protocol import getraddr, LogCallback, Message, message_to_data, moulti_listen

LoopCallback = Callable[[], bool]
MessageCallback = Callable[[Socket, str, Message, FDs], None]
ReadyCallback = Callable[[], None]
SecurityCallback = Callable[[Socket], str]

TYPE_ACCEPT = 1
TYPE_READ_WRITE = 2
TYPE_NOTIFY = 3

class MoultiServer:
	"""
	A server implementation of the Moulti protocol. It uses non-blocking
	sockets polled in an event loop that is expected to run in a dedicated thread.
	"""
	def __init__(
		self,
		instance_name: str|None,
		socket_path: str|None,
		loop_callback: LoopCallback,
		message_callback: MessageCallback,
		log_callback: LogCallback|None = None,
		ready_callback: ReadyCallback|None = None,
		security_callback: SecurityCallback|None = None,
	):
		"""
		socket_path: printable path describing the socket to listen to
		loop_callback: function that returns True if the event loop should keep running, False if it should stop
		message_callback: called after a complete message was received from a client
		log_callback: (optional) called whenever this class wishes to log something
		ready_callback: (optional) called right before the event loop starts; should NOT block
		security_callback: (optional) called after a new client connection was established; should return a non-empty
		error message to deny the connection or an empty string to accept it
		"""
		self.server_selector = DefaultSelector()
		self.instance_name = instance_name or current_instance()
		self.socket_path = socket_path or moulti_socket_path(self.instance_name)
		self.server_socket: Socket|None = None
		self.server_socket_is_abstract = False
		self.loop_callback = loop_callback
		self.message_callback = message_callback
		self.log_callback = log_callback
		self.ready_callback = ready_callback
		self.security_callback = security_callback
		self.replies: Queue = Queue()
		"""Queue used to receive replies that the network loop should send to clients."""
		# That queue cannot be polled through the selectors module: combine it with a Unix pipe:
		self.notifications, self.notify = os.pipe()
		os.set_blocking(self.notifications, False)
		os.set_blocking(self.notify, False)

	def log(self, data: str) -> None:
		if self.log_callback is not None:
			self.log_callback(data)

	def is_listening(self) -> bool:
		return self.server_socket is not None

	def listen(self) -> None:
		self.server_socket, self.server_socket_is_abstract = moulti_listen(bind=from_printable(self.socket_path))
		socket_type = "abstract socket" if self.server_socket_is_abstract else "socket"
		self.log(f'listening on {socket_type} {self.socket_path}')

	def network_loop(self) -> None:
		if not self.is_listening():
			return
		try:
			self.server_selector.register(self.notifications, EVENT_READ, {'type': TYPE_NOTIFY})
			assert self.server_socket is not None
			self.server_selector.register(self.server_socket, EVENT_READ, {'type': TYPE_ACCEPT})

			if self.ready_callback is not None:
				self.ready_callback()

			while self.loop_callback():
				events = self.server_selector.select(0.1)
				# This loop is not entirely event-based: it also dequeues replies, hence the relatively short duration
				# in select().
				for key, mask in events:
					assert isinstance(key.data, dict)
					if key.data['type'] == TYPE_ACCEPT:
						assert isinstance(key.fileobj, Socket)
						self.accept(key.fileobj)
					elif key.data['type'] == TYPE_READ_WRITE:
						if mask & EVENT_WRITE:
							self.write(key.data['writer'])
						if mask & EVENT_READ:
							self.read(key.data['reader'])
					elif key.data['type'] == TYPE_NOTIFY:
						self.clear_notifications()
						self.handle_replies()
		except Exception as exc:
			self.log(f'network loop: {exc}')
		finally:
			clean_socket(self.socket_path)
			# Explicitly stop listening and close all sockets: this makes sense when and if the
			# Python process does not exit after this App has finished running.
			if self.server_socket is not None:
				self.server_socket.close()
				self.server_socket = None
			os.close(self.notify)
			clean_selector(self.server_selector, close_fds=True, close=True)

	def send_notification(self) -> None:
		if not self.is_listening():
			return
		try:
			os.write(self.notify, b'x')
		except BlockingIOError:
			# The pipe is full, but this can be ignored.
			pass

	def clear_notifications(self) -> None:
		try:
			while True:
				os.read(self.notifications, 1024)
		except BlockingIOError:
			# All done: the pipe is empty.
			pass

	def accept(self, socket: Socket) -> None:
		raddr = None
		try:
			connection, _ = socket.accept()
			raddr = getraddr(connection)
			self.log(f'{raddr}: accept: accepted new connection')
			if self.security_callback is not None:
				if error_message := self.security_callback(connection):
					connection.close()
					self.log(f'{raddr}: accept: closed connection: {error_message}')
					return
			connection.setblocking(False)
			data = {
				'type': TYPE_READ_WRITE,
				'reader': MoultiTLVReader(connection, raddr, self.got_tlv, self.log_callback),
				'writer': MoultiTLVWriter(connection, raddr, None, self.log_callback),
			}
			self.server_selector.register(connection, EVENT_READ, data)
		except Exception as exc:
			self.log(f'{raddr}: accept: {exc}')

	def read(self, tlv_reader: MoultiTLVReader) -> None:
		raddr = tlv_reader.raddr
		try:
			try:
				# Read whatever there is to read; if a complete TLV message is received, calls got_tlv, which calls
				# handle_message, which calls reply().
				tlv_reader.read()
			except (MoultiProtocolException, MoultiConnectionClosedException, ConnectionResetError) as exc:
				self.server_selector.unregister(tlv_reader.socket)
				tlv_reader.socket.close()
				self.log(f'{raddr}: read: closed connection; cause: {exc}')
		except Exception as exc:
			self.log(f'{raddr}: read: {exc}')

	def write(self, tlv_writer: MoultiTLVWriter) -> None:
		raddr = tlv_writer.raddr
		try:
			try:
				# Write whatever there is to write:
				all_written = tlv_writer.write()
				self.watch_write_events(tlv_writer.socket, not all_written)
			except (MoultiProtocolException, MoultiConnectionClosedException, ConnectionResetError) as exc:
				self.server_selector.unregister(tlv_writer.socket)
				tlv_writer.socket.close()
				self.log(f'{raddr}: write: closed connection; cause: {exc}')
		except Exception as exc:
			self.log(f'{raddr}: write: {exc}')

	def got_tlv(self, socket: Socket, raddr: str, data_type: str, data: bytes, file_descriptors: FDs) -> None:
		if data_type == 'JSON':
			message = data_to_message(data)
			log_message = abridge_dict(message)
			self.log(f'{raddr}: => message={log_message} {file_descriptors=}')
			self.message_callback(socket, raddr, message, file_descriptors)
		else:
			raise MoultiProtocolException(f'cannot handle {data_type} {len(data)}-byte message')

	def reply(self, socket: Socket, message: Message) -> None:
		# Delegate to the network loop through a queue:
		self.replies.put_nowait((socket, message))
		self.send_notification()
		# This approach makes this function thread-safe.

	def handle_replies(self) -> None:
		try:
			while True:
				socket, message = self.replies.get_nowait()
				self.handle_reply(socket, message)
		except Empty:
			pass

	def handle_reply(self, socket: Socket, message: Message) -> None:
		try:
			# Get the TLV writer for the given socket:
			tlv_writer = self.server_selector.get_key(socket).data['writer']
			# Provide the TLV writer with the reply message so it stores it but does not try to send it:
			tlv_writer.write_message(message_to_data(message), 'JSON', immediate=False)
			# Let write() handle the rest:
			self.write(tlv_writer)
		except Exception as exc:
			self.log(f'handle_reply: {exc}')

	def watch_write_events(self, socket: Socket, watch: bool) -> None:
		"""
		Enable or disable the watching of write events for the given socket.
		Such watching is relevant only when we have data to write to that socket.
		In particular, watching write events for an idle socket leads to 100% CPU consumption as the selector keeps
		returning the socket is ready for writing.
		"""
		key = self.server_selector.get_key(socket)
		events = EVENT_READ
		if watch:
			events |= EVENT_WRITE
		if events != key.events:
			self.server_selector.modify(socket, events, key.data)



from argparse import ArgumentParser, BooleanOptionalAction
from ..abstractstep.cli import add_abstractstep_options

def add_collapsiblestep_options(parser: ArgumentParser, none: bool = False) -> None:
	"""Options common to step add (with actual default values) and step update (with None default values)."""
	add_abstractstep_options(parser, none)
	parser.add_argument('--collapsed', action=BooleanOptionalAction, default=None if none else False, help='whether to collapse the step')
	parser.add_argument('--top-text', '-tt', type=str, default=None if none else '', help='line of text displayed above the content')
	parser.add_argument('--bottom-text', '-bt', type=str, default=None if none else '', help='line of text displayed below the content')



from argparse import ArgumentParser, BooleanOptionalAction, _SubParsersAction
from moulti.client import send_delete, send_to_moulti_and_handle_reply, send_no_none_to_moulti_and_handle_reply
from moulti.environ import pint
from ..abstractquestion.cli import add_abstractquestion_options, question_get_answer

COMMAND = 'inputquestion' # abridged 'iq' below

def add_cli_arguments(subparsers: _SubParsersAction) -> None:
	# moulti inputquestion:
	helpmsg = 'create and manage interactive questions with a single-line input field'
	iq_parser = subparsers.add_parser(COMMAND, help=helpmsg)
	iq_subparsers = iq_parser.add_subparsers(required=True)
	add_iq_commands(iq_subparsers)

def add_iq_commands(iq_subparsers: _SubParsersAction) -> None:
	# moulti inputquestion add
	iq_add_parser = iq_subparsers.add_parser('add', help='add a new input question to Moulti')
	iq_add_parser.set_defaults(func=send_to_moulti_and_handle_reply, command=COMMAND, action='add')
	iq_add_parser.add_argument('id', type=str, help='unique identifier')
	add_abstractquestion_options(iq_add_parser)
	add_inputquestion_options(iq_add_parser)

	# moulti inputquestion update
	iq_update_parser = iq_subparsers.add_parser('update', help='update an existing input question')
	iq_update_parser.set_defaults(func=send_no_none_to_moulti_and_handle_reply, command=COMMAND, action='update')
	iq_update_parser.add_argument('id', type=str, help='unique identifier')
	add_abstractquestion_options(iq_update_parser, none=True)
	add_inputquestion_options(iq_update_parser, none=True)

	# moulti inputquestion get-answer
	helpmsg = 'return the answer provided by the end user'
	iq_getanswer_parser = iq_subparsers.add_parser('get-answer', help=helpmsg)
	iq_getanswer_parser.set_defaults(func=question_get_answer, command=COMMAND, action='get-answer')
	iq_getanswer_parser.add_argument('id', type=str, help='unique identifier')
	iq_getanswer_parser.add_argument('--wait', '-w', action='store_true', help='wait until the answer is available')

	# moulti inputquestion delete
	iq_delete_parser = iq_subparsers.add_parser('delete', help='delete an existing input question')
	iq_delete_parser.set_defaults(func=send_delete, command=COMMAND, action='delete')
	iq_delete_parser.add_argument('id', type=str, nargs='+', help='unique identifier')

def add_inputquestion_options(parser: ArgumentParser, none: bool = False) -> None:
	parser.add_argument('--placeholder', default=None if none else '', help='placeholder')
	parser.add_argument('--value', default=None if none else '', help='value')
	parser.add_argument('--password', default=None if none else False, action=BooleanOptionalAction, help='whether to hide input')
	parser.add_argument('--max-length', default=None if none else 0, type=pint, help='input maximum length (0 to disable)')
	parser.add_argument('--restrict', default=None if none else '', help='Python regular expression to restrict input ("" to disable)')



"""
High-level functions relevant to implement Moulti clients.
"""
import sys
from typing import Any, cast
# Some of the names provided by this module come straight from the pipeline and protocol modules:
# ruff: noqa: F401 imported but unused
# pylint: disable=W0611 unused-import
from .pipeline import pipeline
from .protocol import Message, Socket
from .protocol import moulti_connect, recv_json_message, send_json_message, send_to_moulti
from .protocol import moulti_socket_path, current_instance

Args = dict[str, Any]

def send(socket: Socket, message: Message) -> Message:
	send_json_message(socket, message)
	return recv_json_message(socket, 0)[0]

def handle_reply(reply: Message) -> None:
	success = reply.get('done') is True
	if not success:
		fallback = 'alas, no error message was provided.'
		error_message = reply.get('error', fallback)
		sys.stderr.write(f'Something went wrong: {error_message}\n')
	sys.exit(0 if success else 1)

def send_to_moulti_and_handle_reply(message: Message) -> None:
	reply = cast(Message, send_to_moulti(message))
	handle_reply(reply)

def send_no_none_to_moulti_and_handle_reply(args: Args) -> None:
	no_none_args = {k:v for (k,v) in args.items() if v is not None}
	send_to_moulti_and_handle_reply(no_none_args)

def send_delete(args: Args) -> None:
	errors = pipeline((step_id, {**args, 'id': step_id}, None) for step_id in args['id'])
	sys.exit(1 if errors else 0)



from typing import Any, Iterator, Sequence, cast
from textual import on
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.geometry import Region
from textual.reactive import Reactive
from textual.dom import DOMNode
from textual.widget import AwaitMount, AwaitRemove
from moulti.search import TextSearch
from .vertscroll import VertScroll
from .abstractstep.tui import AbstractStep
from .abstractquestion.tui import AbstractQuestion

class StepContainer(VertScroll):
	"""
	Vertical scrollable container for steps.
	"""
	BINDINGS = [
		Binding('l', 'toggle_scrolling(True)', 'Lock scroll'),
		Binding('l', 'toggle_scrolling(False)', 'Unlock scroll'),
		Binding("ctrl+t", "focus_question(True, False)", "Previous unanswered question", show=False),
		Binding("ctrl+y", "focus_question(False, False)", "Next unanswered question", show=False),
	]

	DEFAULT_CSS = """
	/* Leave a thin space between steps and the container's vertical scrollbar: */
	StepContainer.vertical_scrollbar_visible > Widget {
		margin-right: 1;
	}
	$scrollbar_green: #18b218;
	StepContainer.prevent_programmatic_scrolling {
		scrollbar-color: $scrollbar_green;
		scrollbar-color-active: $scrollbar_green;
		scrollbar-color-hover: $scrollbar_green;
	}
	StepContainer.bottom {
		align-vertical: bottom;
	}
	"""

	def __init__(self, *args: Any, **kwargs: Any):
		super().__init__(*args, **kwargs)
		self.prevent_programmatic_scrolling = False
		self.search_cursor = -1
		"""
		Index of the widget that last returned a positive search result.
		"""

	def action_toggle_scrolling(self, prevent: bool) -> None:
		# Actual toggle:
		self.prevent_programmatic_scrolling = prevent
		# Visual toggle:
		self.refresh_bindings()
		self.set_class(self.prevent_programmatic_scrolling, 'prevent_programmatic_scrolling')

	def check_action(self, action: str, parameters: tuple[object, ...]) -> bool:
		if self.screen.maximized:
			return False
		if action == 'toggle_scrolling':
			return parameters[0] != self.prevent_programmatic_scrolling
		return True

	layout_direction_is_down = Reactive(True)

	def watch_layout_direction_is_down(self, _was_down: bool, is_down: bool) -> None:
		self.sort_children(key=None, reverse=not is_down)

	def ordered_steps(self) -> Iterator[AbstractStep]:
		steps = cast(Sequence[AbstractStep], self.children)
		if self.layout_direction_is_down:
			yield from steps
		else:
			yield from reversed(steps)

	def ordered_index(self, step: AbstractStep) -> int:
		try:
			if self.layout_direction_is_down:
				return self.children.index(step)
			return len(self.children) - 1 - self.children.index(step)
		except ValueError:
			return -1

	def focus_step(self) -> AbstractStep|None:
		try:
			return self.query_one('AbstractStep:focus-within', AbstractStep)
		except NoMatches:
			return None

	def get_question(self, backward: bool = False, answered: bool|None = False) -> AbstractQuestion|None:
		steps = list(self.ordered_steps())
		if backward:
			steps = list(reversed(steps))
		if current := self.focus_step():
			try:
				index = steps.index(current)
				steps = steps[index+1:] + steps[:index]
			except ValueError:
				pass
		for step in steps:
			if isinstance(step, AbstractQuestion):
				if answered is None or answered == step.answered():
					return step
		return None

	def action_focus_question(self, backward: bool = False, answered: bool|None = False) -> None:
		question = self.get_question(backward=backward, answered=answered)
		if question is not None:
			question.focus_input_widget()

	def add_step(self, step: AbstractStep) -> AwaitMount:
		return self.mount(step, before=None if self.layout_direction_is_down else 0)

	def remove_step(self, step: AbstractStep) -> AwaitRemove|None:
		deleted_index = self.ordered_index(step)
		if deleted_index != -1:
			removal = step.remove()
			if deleted_index <= self.search_cursor:
				self.search_cursor -= 1
			return removal
		return None

	def parent_step(self, widget: DOMNode|None) -> AbstractStep|None:
		while widget is not None and not isinstance(widget, AbstractStep):
			widget = widget.parent
		return widget

	def search_maximized(self, search: TextSearch) -> bool:
		# Is there a maximized widget?
		maximized_widget = self.screen.maximized
		if not maximized_widget:
			return False
		# Does the maximized widget belong to a step?
		step = self.parent_step(maximized_widget)
		if step is None:
			return False
		# Does that step belong to this container?
		if step.parent != self:
			return False
		# Can this step handle the matter?
		if hasattr(step, 'search_maximized') and step.search_maximized(search):
			# A match was found inside the maximized widget:
			# 1 - discard the previously highlighted match, if any:
			step_index = self.ordered_index(step)
			if self.search_cursor not in (-1, step_index):
				steps = list(self.ordered_steps())
				current_step = steps[self.search_cursor]
				if hasattr(current_step, 'search'):
					current_step.search(TextSearch.make_reset())
			# 2 - update the search cursor so as to resume search from the maximized widget's parent step:
			self.search_cursor = step_index
			return True
		return False

	def search(self, search: TextSearch) -> bool:
		if self.screen.maximized:
			return self.search_maximized(search)
		steps = list(enumerate(self.ordered_steps()))
		if self.search_cursor != -1:
			steps = steps[self.search_cursor:] if search.next_result else steps[:self.search_cursor+1]
		iter_steps = iter(steps) if search.next_result else reversed(steps)
		for index, widget in iter_steps:
			if not hasattr(widget, 'search'):
				continue
			if widget.search(search):
				self.search_cursor = index
				return True
		self.search_cursor = -1
		return False

	def scroll_to_step(self, step: AbstractStep, where: bool|int = True) -> None:
		if self.prevent_programmatic_scrolling or where is False:
			return
		if where is True:
			self.scroll_to_widget(step, origin_visible=True)
			return
		step_region = step.virtual_region
		line = (step_region.bottom if where < 0 else step_region.y) + where
		target_region = Region(step_region.x, line, step_region.width, 1)
		self.scroll_to_region(target_region)

	@on(AbstractStep.StepActivity)
	def on_step_activity(self, activity: AbstractStep.StepActivity) -> None:
		# Consume the event:
		activity.prevent_default()
		activity.stop()
		# Extract step and its scroll policy:
		step = activity.step
		scroll_on_activity = step.scroll_on_activity
		# Delay the actual scrolling:
		if scroll_on_activity is not False:
			def do_scroll() -> None:
				self.scroll_to_step(step, scroll_on_activity)
			self.set_timer(0.05, do_scroll)

	@on(AbstractStep.ScrollRequest)
	def on_scroll_request(self, request: AbstractStep.ScrollRequest) -> None:
		# Consume the event:
		request.prevent_default()
		request.stop()
		if isinstance(request.target, Region):
			# The provided region must be mapped to this container's coordinates:
			region = request.target # relative to the AbstractStep widget
			region = region.translate(request.step.virtual_region.offset) # relative to this StepContainer
			self.scroll_to_region(region, **request.scroll_kwargs)
			return
		target_widget = request.step if request.target is None else request.target
		self.scroll_to_widget(target_widget, **request.scroll_kwargs)



from typing import Any, Iterable
from typing_extensions import Self
from rich.text import Text
from textual.color import Color
from textual.events import MouseScrollUp, Click
from textual.geometry import Region, Size, Spacing
from textual.scroll_view import ScrollView
from textual.strip import Strip
from moulti.search import MatchSpan, TextSearch
from moulti.widgets.mixin import ToLinesMixin

class MoultiLog(ScrollView, ToLinesMixin, can_focus=True):
	"""
	This widget is a lightweight variant of RichLog with:
	- the ability to use the App's ANSI theme (background and foreground)
	- custom scrolling behaviour
	- a focus indicator on the left
	"""
	BINDINGS = [
		("f", "maximize", "Maximize"),
		("f", "minimize", "Exit maximized mode"),
	]

	def __init__(
		self,
		*,
		auto_scroll: bool = True,
		name: str | None = None,
		id: str | None = None,  # pylint: disable=redefined-builtin
		classes: str | None = None,
		disabled: bool = False,
	) -> None:
		super().__init__(name=name, id=id, classes=classes, disabled=disabled)
		self.lines: list[str|Text|Strip] = []
		self.max_width: int = 0
		self.auto_scroll = self.default_auto_scroll = auto_scroll
		self.follow_ansi_theme = True
		self.search_cursor: tuple[int|None, MatchSpan] = (None, None)
		self.searched_line_backup: tuple[int, str|Text|Strip] = (-1, '')
		self.max_height_backup: Any = None

	def set_max_height(self, new_max_height: int|None) -> int|None:
		if new_max_height is None:
			actual_max_height = new_max_height
		else:
			actual_max_height = new_max_height if new_max_height > 0 else None

		if self.is_maximized:
			# Maximized mode should ignore `moulti step update id --max-height=42`:
			self.max_height_backup = actual_max_height
		else:
			self.styles.max_height = actual_max_height

		return actual_max_height

	def action_maximize(self) -> None:
		self.max_height_backup = self.styles.max_height
		self.styles.max_height = '99h' # Leave some room for the search widget
		self.screen.maximize(self, False)
		self.refresh_bindings()

	def action_minimize(self) -> None:
		self.screen.minimize()
		self.styles.max_height = self.max_height_backup
		self.refresh_bindings()

	def check_action(self, action: str, parameters: tuple[object, ...]) -> bool:
		if action == 'maximize':
			return not bool(self.screen.maximized)
		if action == 'minimize':
			return bool(self.screen.maximized)
		return True

	def clear_search(self) -> None:
		self.search_cursor = (None, None)
		self.searched_line_backup = (-1, '')

	def on_mount(self) -> None:
		if self.follow_ansi_theme:
			self.watch(self.app, 'dark', self.apply_ansi_fgbg)

	def on_unmount(self) -> None:
		if self.is_maximized:
			self.action_minimize()

	def apply_ansi_fgbg(self, _old: bool, _new: bool) -> None:
		theme = self.app.ansi_theme
		self.styles.auto_color = False
		self.styles.color = Color(*theme.foreground_color)
		self.styles.background = Color(*theme.background_color)

	def watch_scroll_y(self, old_value: float, new_value: float) -> None:
		"""
		Adjust the behavior of the vertical scrollbar so users can scroll freely despite incoming new lines.
		"""
		# If users grab the vertical scrollbar, auto_scroll should turn off:
		if self.auto_scroll and self.is_vertical_scrollbar_grabbed:
			self.auto_scroll = False
		# ScrollView also implements this watch method: call it or it breaks the entire scrolling.
		super().watch_scroll_y(old_value, new_value)

	def on_mouse_scroll_up(self, _: MouseScrollUp) -> None:
		"""Turn auto_scroll off as soon as users scroll up using the mouse wheel."""
		self.auto_scroll = False

	def action_scroll_up(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll off as soon as users hit the Up key."""
		self.auto_scroll = False
		super().action_scroll_up(*args, **kwargs)

	def action_page_up(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll off as soon as users hit the PgUp key."""
		self.auto_scroll = False
		super().action_page_up(*args, **kwargs)

	def action_scroll_home(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll off as soon as users hit the Home key."""
		self.auto_scroll = False
		super().action_scroll_home(*args, **kwargs)

	def action_scroll_end(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll on again as soon as users hit the End key."""
		self.auto_scroll = self.default_auto_scroll
		super().action_scroll_end(*args, **kwargs)

	def clear(self) -> Self:
		self.clear_search()
		self.lines.clear()
		self.max_width = 0
		self.virtual_size = Size(self.max_width, len(self.lines))
		self.auto_scroll = self.default_auto_scroll
		self.refresh()
		return self

	async def on_click(self, event: Click) -> None:
		# Clicks grant focus to this widget; keep them.
		# See also: AbstractStep.on_click().
		event.stop()

	def write_lines(self, lines: Iterable[str|Text|Strip], max_cell_len: int) -> Self:
		if max_cell_len > self.max_width: # pylint: disable=consider-using-max-builtin
			self.max_width = max_cell_len
		self.lines.extend(lines)

		self.virtual_size = Size(self.max_width, len(self.lines))
		if self.auto_scroll:
			self.scroll_end(animate=False)

		return self

	def render_line(self, y: int) -> Strip:
		scroll_x, scroll_y = self.scroll_offset
		if scroll_y + y >= len(self.lines):
			return Strip.blank(self.size.width, self.rich_style)
		line = self.line_to_strip(scroll_y + y, True)
		line = line.crop_extend(scroll_x, scroll_x + self.size.width, self.rich_style)
		return line.apply_style(self.rich_style)

	def backup_line(self, line_number: int) -> None:
		line = self.lines[line_number]
		# str and Strip are immutable but Text is mutable and thus must be explicitly copied:
		backup_line = line.copy() if isinstance(line, Text) else line
		self.searched_line_backup = (line_number, backup_line)

	def restore_line(self, line_number: int) -> None:
		backup_line_number, backup_line = self.searched_line_backup
		if line_number != backup_line_number:
			return
		restore_line = backup_line.copy() if isinstance(backup_line, Text) else backup_line
		self.lines[line_number] = restore_line

	def search_one_line(self, search: TextSearch) -> bool:
		line_number, highlight = self.search_cursor
		assert line_number is not None
		first_occurrence_on_this_line = highlight is None
		self.restore_line(line_number)
		plain_text, text = self.line_to_plain_text(line_number, True)
		result = search.search(plain_text, highlight)
		if result is None:
			self.restore_line(line_number)
			self.refresh()
			return False
		# Occurrence found!
		self.search_cursor = (line_number, result)
		# Highlight:
		if text is None:
			text = self.line_to_text(line_number, True)
		if first_occurrence_on_this_line:
			self.backup_line(line_number)
		self.lines[line_number] = search.highlight(text, *result)
		return True

	def search(self, search: TextSearch) -> bool:
		if search.reset():
			if (line_number := self.search_cursor[0]) is not None:
				self.restore_line(line_number)
				self.refresh()
				self.clear_search()
			return False
		if search.next_result:
			first_line, next_line = 0, 1
		else:
			first_line, next_line = len(self.lines) - 1, -1
		if self.search_cursor[0] is None:
			self.search_cursor = (first_line, None)
		line_number = self.search_cursor[0]
		assert line_number is not None
		while 0 <= line_number < len(self.lines):
			if self.search_one_line(search):
				return True
			line_number += next_line
			self.search_cursor = (line_number, None)
		self.clear_search()
		return False

	def scroll_to_search_highlight(self) -> tuple[Region, Spacing]|None:
		line_number, highlight = self.search_cursor
		if line_number is None or highlight is None:
			return None
		highlight_region = Region(
			highlight[0],
			line_number,
			highlight[1] - highlight[0],
			1
		)
		highlight_spacing = Spacing(1, 5, 1, 10)
		self.auto_scroll = False
		self.scroll_to_region(highlight_region, spacing=highlight_spacing, animate=False)
		self.refresh()
		# Return the highlighted region relatively to this widget:
		highlight_region = highlight_region.translate(-self.scroll_offset)
		return highlight_region, highlight_spacing

	DEFAULT_CSS = """
	MoultiLog {
		background: $surface;
		color: $text;
		overflow-y: scroll;
		height: auto;
		border-left: blank;
	}
	MoultiLog:focus {
		border-left: thick $primary-lighten-3;
	}
	"""



from moulti.search import TextSearch
from ..abstractstep.tui import AbstractStep

class Divider(AbstractStep, can_focus=True):
	"""
	Step that simply displays its title.
	"""

	BINDINGS = [
		("c", "to_clipboard", "Copy"),
	]

	def on_mount(self) -> None:
		self.tooltip = f'Step id: {self.title_from_id()}'

	def update_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		super().update_properties(kwargs)
		self.update(self.title)

	@AbstractStep.copy_to_clipboard
	def action_to_clipboard(self) -> tuple[bool, str, str]:
		lines_count = len(self.title.split('\n'))
		return True, self.title, f'copied {lines_count} lines, {len(self.title)} characters to clipboard'

	def search(self, search: TextSearch) -> bool:
		found, text = self.search_label(self.title, search)
		self.update(text)
		if found:
			self.focus(False)
			self.post_message(self.ScrollRequest(self, None, center=True, animate=False))
		return found

	DEFAULT_CSS = AbstractStep.DEFAULT_CSS + """
	Divider {
		text-align: center;
		border-top: hkey yellow;
		border-bottom: hkey yellow;
	}
	"""
MoultiWidgetClass = Divider



"""
This file can be used to run instructions early, before the interpreter tries to import Moulti modules.
"""
# PYTHON_ARGCOMPLETE_OK

import sys
if sys.platform == 'win32':
	WINDOWS_MESSAGE = """It seems you are trying to run moulti on Microsoft Windows.
Alas, as these lines are being written:
- Moulti does not work on this operating system.
- The blocking point is https://github.com/python/cpython/issues/77589

IF this blocking point appears to be resolved, try to upgrade Moulti.
If you are still getting this message, feel free to report this issue.
"""
	print(WINDOWS_MESSAGE)
	sys.exit(1)

from .cli import main # pylint: disable=wrong-import-position; multiple import statements fail on Microsoft Windows
assert bool(main) # main will be called by the console_scripts wrapper; this makes all linters happy.



"""
Provide helpers to assign human-friendly names to system threads.
"""

import sys
from ctypes import CDLL, c_char_p, c_void_p
from ctypes.util import find_library
from functools import cache, wraps
from threading import current_thread
from typing import Any, Callable
from .environ import env
from .helpers import abridge_string

def thread_name_max_len() -> int:
	"""
	Return the maximum name length (excluding trailing null byte) for system threads.
	"""
	platform = sys.platform
	# Most Unix operating systems truncate thread names to MAXCOMLEN characters + a null byte:
	max_len = 15
	if platform.startswith("darwin"):
		max_len = 63
	elif platform.startswith("freebsd"):
		max_len = 19
	elif platform.startswith("openbsd"):
		max_len = 23
	elif platform.startswith("netbsd"):
		max_len = 15
	return max_len


THREAD_NAME_MAX_LEN = thread_name_max_len()


@cache
def pthread_set_name_function() -> Callable:
	"""
	Return a simple wrapper around pthread_setname_np().
	This wrapper expects a name as single string (str) argument and assigns it to the current thread.
	"""
	# Access the pthread library:
	pthread_lib = CDLL(find_library("pthread"))

	# Access the pthread_self function:
	pthread_self = pthread_lib.pthread_self
	pthread_self.restype = c_void_p

	# Access the pthread_set_name_np function:
	try:
		pthread_setname_np = pthread_lib.pthread_setname_np
	except AttributeError:
		pthread_setname_np = pthread_lib.pthread_set_name_np

	# Deal with its non-portable (hence "_np") arguments:
	if sys.platform.startswith("netbsd"):
		pthread_setname_np.argtypes = [c_void_p, c_char_p, c_char_p]
		def make_args(name: bytes) -> tuple:
			return (pthread_self(), b"%s\0", name)
	elif sys.platform.startswith("darwin"):
		pthread_setname_np.argtypes = [c_char_p]
		def make_args(name: bytes) -> tuple:
			return (name,)
	else:
		pthread_setname_np.argtypes = [c_void_p, c_char_p]
		def make_args(name: bytes) -> tuple:
			return (pthread_self(), name,)

	def function(name: str) -> None:
		name_bytes = name.encode("ascii", "replace")[:THREAD_NAME_MAX_LEN] + b"\0"
		pthread_setname_np(*make_args(name_bytes))
	return function


def set_thread_name(name: str) -> None:
	"""
	Assign the given name to the current thread.
	If necessary, the name gets abridged so as to fit system limitations (typically 15 or 63 characters max).
	"""
	if env("MOULTI_NAME_THREADS") == "no":
		return
	abridged_name = abridge_string(name, threshold=THREAD_NAME_MAX_LEN, sep="")

	# Python >= 3.14 assigns names to system threads out of the box:
	if sys.version_info >= (3, 14):
		current_thread().name = abridged_name
	else: # Previous versions of Python:
		try:
			pthread_set_name_np = pthread_set_name_function()
			pthread_set_name_np(abridged_name)
		except Exception:
			# Best-effort: we are not even interested in what went wrong.
			pass



IDLE_THREAD_NAME = "thread-pool"
"""
Thread name set by the thread_name() decorator after a decorated function
returns. This value reflects the underlying thread has become idle and has
reintegrated the thread pool.
"""

KEEP_THREAD_NAME = None
"""
Legible constant, to be used with the thread_name() decorator.
"""

def thread_name(start: str|None = KEEP_THREAD_NAME, end: str|None = IDLE_THREAD_NAME) -> Callable:
	"""
	Decorator that calls:
	- set_thread_name(start) before the decorated function, unless start is KEEP_THREAD_NAME;
	- set_thread_name(end) after the decorated function, unless end is KEEP_THREAD_NAME.
	"""
	def thread_name_decorator(func: Callable) -> Callable:
		@wraps(func)
		def wrapper(*args: Any, **kwargs: Any) -> Any:
			try:
				if start is not KEEP_THREAD_NAME:
					set_thread_name(start)
				return func(*args, **kwargs)
			finally:
				if end is not KEEP_THREAD_NAME:
					set_thread_name(end)
		return wrapper
	return thread_name_decorator



from typing import Any
from typing_extensions import Self
from rich.highlighter import Highlighter, RegexHighlighter, ReprHighlighter
from rich.text import Text
from rich.theme import Theme
from textual.app import ComposeResult
from textual.events import MouseScrollUp, Click
from textual.widgets import Label, Log, Static
from moulti.clipboard import copy
from moulti.widgets.mixin import ToLinesMixin

MOULTI_THEME = Theme({
	'moulti.date': 'orchid2',
	'moulti.time': 'orchid2',
	'moulti.socket': 'steel_blue1',
	'moulti.fileno': 'orchid1',
	'moulti.function': 'slate_blue1',
	'moulti.messagein': 'bold dodger_blue2',
	'moulti.messageout': 'bold chartreuse3',
	# Redefine some repr.* styles to avoid ANSI colors:
	'repr.bool_true': 'green1',
	'repr.bool_false': 'deep_pink2',
	'repr.none': 'purple',
	'repr.number': 'cornflower_blue',
	'repr.str': 'sea_green3',
	'repr.attrib_name': 'dark_orange3',
})


class MoultiConsole(Static):
	BINDINGS = [
		("c", "to_clipboard(False)", "Copy"),
		("w", "to_clipboard(True)", "With colors"),
	]

	def __init__(self, *args: Any, **kwargs: Any) -> None:
		super().__init__(*args, **kwargs)
		self.log_widget = MoultiConsoleLog()
		self.log_widget.highlighter = MoultiConsoleHighlighter()
		self.log_widget.highlight = True

	def compose(self) -> ComposeResult:
		self.app.console.push_theme(MOULTI_THEME)
		yield Label('Moulti console')
		yield self.log_widget

	def write(self, content: str) -> Self:
		self.log_widget.write(content)
		return self

	async def on_click(self, event: Click) -> None:
		# Clicking the console (label or log) means focusing its log widget:
		self.log_widget.focus()
		event.stop()

	async def action_to_clipboard(self, keep_styles: bool = False) -> None:
		try:
			copy(self.app, '\n'.join(self.log_widget.to_lines(keep_styles)))
			title, message, res = 'Console copied!', 'Console lines copied to clipboard', True
		except Exception as exc:
			title, message, res = 'Failed to copy console!', str(exc), False
		self.app.notify(message, title=title, severity='information' if res else 'error')

	DEFAULT_CSS = """
	$moulti_console_color: $primary;
	MoultiConsole {
		Label {
			width: 100%;
			background: $moulti_console_color;
			color: auto;
		}
		MoultiConsoleLog {
			min-height: 5;
			max-height: 15;
			border-bottom: none;
			border-top: none;
			border-right: thick $moulti_console_color;
			border-left: thick $moulti_console_color;
			scrollbar-corner-color: $moulti_console_color;
		}
		&.hidden {
			display: none;
		}
	}
	"""

class MoultiConsoleLog(Log, ToLinesMixin):
	"""
	Moulti's console leverages Log and its highlighter
	"""
	def watch_scroll_y(self, old_value: float, new_value: float) -> None:
		"""
		Adjust the behavior of the vertical scrollbar so users can scroll freely despite incoming new lines.
		"""
		# If users grab the vertical scrollbar, auto_scroll should turn off:
		if self.auto_scroll and self.is_vertical_scrollbar_grabbed:
			self.auto_scroll = False
		# ScrollView also implements this watch method: call it or it breaks the entire scrolling.
		super().watch_scroll_y(old_value, new_value)

	# pylint: disable=duplicate-code; copy/paste so as to get the same scrolling behaviour as MoultiLog.
	# Moving duplicate code to a common Mixin works but does not please mypy.
	def on_mouse_scroll_up(self, _: MouseScrollUp) -> None:
		"""Turn auto_scroll off as soon as users scroll up using the mouse wheel."""
		self.auto_scroll = False

	def action_scroll_up(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll off as soon as users hit the Up key."""
		self.auto_scroll = False
		super().action_scroll_up(*args, **kwargs)

	def action_page_up(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll off as soon as users hit the PgUp key."""
		self.auto_scroll = False
		super().action_page_up(*args, **kwargs)

	def action_scroll_home(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll off as soon as users hit the Home key."""
		self.auto_scroll = False
		super().action_scroll_home(*args, **kwargs)

	def action_scroll_end(self, *args: Any, **kwargs: Any) -> None:
		"""Turn auto_scroll on again as soon as users hit the End key."""
		self.auto_scroll = True
		super().action_scroll_end(*args, **kwargs)

	DEFAULT_CSS = """
	MoultiConsoleLog {
		height: auto;
		border-left: blank;
		background: $background;
	}
	MoultiConsoleLog:focus {
		border-left: thick $primary-lighten-3;
		background-tint: white 0%;
	}
	"""

class MoultiRegexHighlighter(RegexHighlighter):
	RE_EXT_MODE = '(?x) # extended mode'
	DATE = r'''
(?P<date>
	[1-9][0-9]*-           # year
	(?:0[1-9]|1[0-2])-     # month
	(?:[012][0-9]|3[01])   # day
)'''
	TIME = r'''
(?P<time>
	(?:[01][0-9]|2[0-3]):  # hours
	(?:[0-5][0-9]):        # minutes
	(?:[0-5][0-9])\.       # seconds
	[0-9]+                 # milliseconds
)'''
	DATETIME = DATE + '[T ]' + TIME
	SOCKET = r'(?P<socket>[@/].*\.socket)'
	FILENO = r'fd(?P<fileno>\d+)'
	SOCKET_FILENO = SOCKET + ':' + FILENO
	FUNCTION = r'(?P<function>accept|exec|network loop|pass|read|reply|save)'
	MESSAGE_IN = r'(?P<messagein>=>)'
	MESSAGE_OUT = r'(?P<messageout><=)'
	MESSAGE = '(?:' + MESSAGE_IN + '|' + MESSAGE_OUT + ')'

	base_style: str = "moulti."
	highlights = [
		RE_EXT_MODE + '^' + DATETIME + ' ',
		' ' + SOCKET_FILENO + ': ',
		SOCKET,
		' ' + FUNCTION + ': ',
		': ' + MESSAGE + ' ',
	]

class MoultiConsoleHighlighter(Highlighter):
	SEPARATORS = (' => ', ' <= ', ' exec: ', ' theme policy: ', ' quit policy: ')
	"""
	Everything to the right of these separators is to be highlighted by ReprHighlighter.
	Everything to the left of these separators is to be highlighted by MoultiRegexHighlighter.
	"""

	def __init__(self) -> None:
		self.repr_hl = ReprHighlighter()
		self.moulti_re_hl = MoultiRegexHighlighter()
		super().__init__()

	def highlight(self, text: Text) -> None:
		plain_text = text.plain
		for separator in self.SEPARATORS:
			if separator in plain_text:
				left, right = plain_text.split(separator, 1)
				# Highlight the left part (including the separator):
				text.set_length(len(left) + len(separator))
				self.moulti_re_hl.highlight(text)
				# Highlight the right part:
				right_text = Text(right)
				self.repr_hl.highlight(right_text)
				text.append_text(right_text)
				# All done:
				return
		# No separator? Highlight the entire string using MoultiRegexHighlighter:
		self.moulti_re_hl.highlight(text)



from .common import moulti_test, moulti_test_steps
assert moulti_test

def test_divider(moulti_test):
	"""
	Run all commands taken from the documentation's "Dividers" section.
	"""
	moulti_test_steps(moulti_test, 'divider.bash', 3)



import pytest
from moulti.cli import adjust_cli_args, first_non_option_argument, inject_double_dash_before_command

def test_first_non_option_argument():
	assert first_non_option_argument(['--foo', '-b']) is None
	assert first_non_option_argument(['foo']) == 0
	assert first_non_option_argument(['--foo', '-b', 'bar']) == 2
	assert first_non_option_argument(['moulti', 'run', 'foo'], 2) == 2
	assert first_non_option_argument(['moulti', 'run', '--foo', '-b', 'foo'], 2) == 4

INJECT_DOUBLE_DASH_BEFORE_COMMAND = [
	{
		'args': ['--foo', '-b'],
		'start': 0,
		'result': ['--foo', '-b'],
	},
	{
		'args': ['foo'],
		'start': 0,
		'result': ['foo'],
	},
	{
		'args': ['--foo', '-b', 'bar'],
		'start': 0,
		'result': ['--foo', '-b', '--', 'bar'],
	},
	{
		'args': ['moulti', 'run', 'foo'],
		'start': 2,
		'result': ['moulti', 'run', '--', 'foo'],
	},
	{
		'args': ['moulti', 'run', '--foo', '-b', 'foo'],
		'start': 2,
		'result': ['moulti', 'run', '--foo', '-b', '--', 'foo'],
	},
	{
		'args': ['moulti', 'run', '--foo', '-b', '--', 'foo'],
		'start': 2,
		'result': ['moulti', 'run', '--foo', '-b', '--', 'foo'],
	},
]

@pytest.mark.parametrize('test', INJECT_DOUBLE_DASH_BEFORE_COMMAND)
def test_inject_double_dash_before_command(test):
	inject_double_dash_before_command(test['args'], test['start'])
	assert test['args'] == test['result']

ADJUST_CLI_ARGS = [
	{'args': ['moulti', 'run']},
	{'args': ['moulti', 'run', '--help']},
	{'args': ['moulti', 'run', 'ls'], 'result': ['moulti', 'run', '--', 'ls']},
	{'args': ['moulti', 'run', 'ls', '-al'], 'result': ['moulti', 'run', '--', 'ls', '-al']},
	{
		'args': ['moulti', 'run', 'moulti', 'diff', 'run', 'git', 'diff', '--cached'],
		'result': ['moulti', 'run', '--', 'moulti', 'diff', 'run', 'git', 'diff', '--cached'],
	},
	{
		'args': ['moulti', 'diff', 'run', 'git', 'diff'],
		'result': ['moulti', 'diff', 'run', '--', 'git', 'diff'],
	},
	{
		'args': ['moulti', 'diff', 'run', 'git', 'diff', '--cached'],
		'result': ['moulti', 'diff', 'run', '--', 'git', 'diff', '--cached'],
	},
]

@pytest.mark.parametrize('test', ADJUST_CLI_ARGS)
def test_adjust_cli_args(test):
	expectation = test.get('result') or test['args'].copy()
	adjust_cli_args(test['args'])
	assert test['args'] == expectation



from typing import Any
from textual import work
from textual.app import ComposeResult
from textual.message import Message
from textual.widget import Widget
from textual.widgets import Static
from moulti.search import TextSearch
from ..collapsiblestep.tui import CollapsibleStep

class AbstractQuestion(CollapsibleStep):
	"""
	This widget represents an interactive question in a script, program or process.
	"""
	def __init__(self, id: str, **kwargs: str|int|bool): # pylint: disable=redefined-builtin
		self.question = ''
		self.question_label = Static(self.question, classes='question_text')
		self.answer: str|None = None
		self.waiting: list[Any] = []
		super().__init__(id=id, **kwargs)

	def cli_action_get_answer(self, kwargs: dict[str, str|int|bool], helpers: dict[str, Any]) -> tuple:
		if self.answer is None:
			if kwargs.get('wait'):
				self.waiting.append((kwargs, helpers))
			else:
				helpers['reply'](done=False, error='answer not available yet')
		else:
			helpers['reply'](answer=self.answer, done=True, error=None)
		return ()

	@work(thread=True, group='question-flush', name='flush-answers')
	def flush_waiting_clients(self) -> None:
		for _kwargs, helpers in self.waiting:
			try:
				if self.answer is None:
					helpers['reply'](done=False, error='no answer from end user')
				else:
					helpers['reply'](done=True, error=None, answer=self.answer)
			except Exception as exc:
				error = str(exc)
				helpers['debug'](f'flush_waiting_clients: {error}')
		self.waiting.clear()

	def subcompose(self) -> ComposeResult:
		yield self.question_label
		yield from self.compose_question()

	def subsearch(self, search: TextSearch) -> bool:
		found, text = self.search_label(self.question, search)
		self.question_label.update(text)
		if found:
			self.focus(False)
			self.post_message(self.ScrollRequest(self, self.question_label, center=True, animate=False))
		return found

	def compose_question(self) -> ComposeResult:
		return []

	def check_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		super().check_properties(kwargs)
		self.check_markup_dict(kwargs, 'text')

	def update_properties(self, kwargs: dict[str, str|int|bool]) -> None:
		super().update_properties(kwargs)
		if 'text' in kwargs:
			self.question = str(kwargs['text'])
			self.question_label.update(self.question)

	def export_properties(self) -> dict[str, Any]:
		prop = super().export_properties()
		prop['text'] = str(self.question_label.renderable)
		return prop

	def disable(self) -> None:
		for widget in self.collapsible.query('Collapsible > Contents Widget').results(Widget):
			if widget.focusable:
				widget.blur()
				widget.disabled = True

	def focus_input_widget(self) -> None:
		self.collapsible.collapsed = False
		for widget in self.collapsible.query('Collapsible > Contents Widget').results(Widget):
			if widget.focusable:
				widget.focus()
				break

	def got_answer(self, answer: str) -> None:
		# Disable the question to ensure users answer only once:
		self.disable()
		# Store the answer for future retrieval:
		self.answer = answer
		self.flush_waiting_clients()
		# Emit a GotAnswer message so other components react to the event:
		self.post_message(GotAnswer(self, self.answer))

	def answered(self) -> bool:
		return self.answer is not None

class GotAnswer(Message):
	def __init__(self, origin: AbstractQuestion, answer: str) -> None:
		self.origin = origin
		self.answer = answer
		super().__init__()



import re
from collections import deque
from rich.text import Text

# Default value for endpos in re.Pattern.search():
# re.Pattern.search(self, string, pos=0, endpos=9223372036854775807)
END = 2**63-1

MatchSpan = tuple[int, int]|None

class TextSearch:
	def __init__(self, pattern: str, regex: bool, case_insensitive: bool, next_result: bool, reset: bool = False) -> None:
		self.pattern = pattern
		self.regex = regex
		self.case_insensitive = case_insensitive
		self.next_result = next_result
		self.is_reset = reset
		if self.regex:
			re_flags = re.IGNORECASE if self.case_insensitive else 0
			self._pattern = re.compile(self.pattern, re_flags)
		elif self.case_insensitive:
			self._casefold_pattern = self.pattern.casefold()

	def __str__(self) -> str:
		if self.is_reset:
			return 'reset text search'
		pattern_type = 'regex' if self.regex else 'basic'
		cs_type = 'case-insensitive' if self.case_insensitive else 'case-sensitive'
		direction = 'Forward' if self.next_result else 'Backward'
		return f'{direction} text search: {pattern_type} pattern "{self.pattern}", {cs_type}'

	def __repr__(self) -> str:
		return self.__str__()

	def copy(self) -> 'TextSearch':
		return TextSearch(self.pattern, self.regex, self.case_insensitive, self.next_result)

	@classmethod
	def make_reset(cls) -> 'TextSearch':
		return cls('', False, False, False, True)

	def reset(self) -> bool:
		return self.is_reset

	def search(self, plain_text: str, highlight: MatchSpan) -> MatchSpan:
		if not plain_text or self.is_reset:
			return None
		if self.regex:
			return self.regex_search(plain_text, highlight)
		return self.simple_search(plain_text, highlight)

	def regex_search(self, plain_text: str, highlight: MatchSpan) -> MatchSpan:
		start, end = (None, None) if highlight is None else highlight
		rem = None
		if self.next_result:
			# First match after the end of the highlighted part:
			rem = self._pattern.search(plain_text, end or 0)
		else:
			# Last match before the start of the highlighted part:
			matches = self._pattern.finditer(plain_text, 0, END if start is None else start)
			if last_match := deque(matches, maxlen=1):
				rem = last_match.pop()
		if rem:
			return rem.span()
		return None

	def simple_search(self, plain_text: str, highlight: MatchSpan) -> MatchSpan:
		pattern = self.pattern
		start, end = (None, None) if highlight is None else highlight
		if self.case_insensitive:
			plain_text = plain_text.casefold()
			pattern = self._casefold_pattern
		if self.next_result:
			start = plain_text.find(pattern, end)
		else:
			start = plain_text.rfind(pattern, 0, start)
		if start != -1:
			return (start, start + len(pattern))
		return None

	def highlight(self, text: Text, start: int, end: int) -> Text:
		text.stylize('reverse', start, end)
		return text



from .common import moulti_test, moulti_test_steps
assert moulti_test

def test_first_steps(moulti_test):
	"""
	Run all commands taken from the documentation's "First steps with Moulti" section.
	"""
	moulti_test_steps(moulti_test, 'first-steps.bash', 12)



import errno
import os
import re
import sys
import pwd
import uuid
from collections import deque
from socket import socket as Socket, AF_UNIX, SOCK_STREAM
from socket import recv_fds as socket_recv_fds, send_fds
import json
from json.decoder import JSONDecodeError
from typing import Any, Callable

Message = dict[str, Any]
FDs = list[int]
TLVCallback = Callable[[Socket, str, str, bytes, FDs], None]
LogCallback = Callable[[str], None]

# recv_fds() uses CMSG_LEN instead of CMSG_SPACE to compute the ancillary data buffer size (socklen_t msg_controllen)
# and is thus liable to underestimate it. On some platforms (e.g. FreeBSD+amd64), this results in MSG_CTRUNC.
# A simple workaround is to adjust/increment max_fds.
CMSG_ADJUSTMENT = 1 if 'bsd' in sys.platform else 0

def recv_fds(sock: Socket, bufsize: int, maxfds: int, flags: int = 0) -> tuple[Any, Any, Any, Any]:
	return socket_recv_fds(sock, bufsize, maxfds + CMSG_ADJUSTMENT, flags)

def abstract_unix_sockets_supported() -> bool:
	# According to https://github.com/microsoft/WSL/issues/4240#issuecomment-620805115 :
	# > "Windows implementation of AF_UNIX does not support abstract sockets"
	# Consequently, assume only Linux supports abstract Unix sockets:
	return sys.platform == 'linux'

def run_dir() -> str:
	# This mess is *exactly* why we prefer abstract Unix sockets:
	if 'XDG_RUNTIME_DIR' in os.environ:
		return os.environ['XDG_RUNTIME_DIR']
	our_directory = 'moulti'
	if 'XDG_CACHE_HOME' in os.environ:
		return os.path.join(os.environ['XDG_CACHE_HOME'], our_directory)
	caches = os.path.expanduser('~/Library/Caches')
	if os.path.isdir(caches):
		return os.path.join(caches, our_directory)
	return os.path.join(os.path.expanduser('~/.cache'), our_directory)
	# Fallback on /tmp would have brought security implications (e.g. umask).

def current_username() -> str:
	uid = os.getuid()
	return pwd.getpwuid(uid).pw_name

def current_instance() -> str:
	return os.environ.get('MOULTI_INSTANCE') or 'default'

def make_socket_path(filename: str, abstract: bool, dirpath: str|None = None, printable: bool = True) -> str:
	if abstract:
		return ('@' if printable else '\0') + filename
	return os.path.join(dirpath or run_dir(), filename)

def default_moulti_socket_path(instance: str|None = None) -> str:
	username = current_username()
	instance = instance or current_instance()
	socket_filename = f'moulti-{username}-{instance}.socket'
	return make_socket_path(socket_filename, abstract_unix_sockets_supported())

def moulti_socket_path(instance: str|None = None) -> str:
	return os.environ.get('MOULTI_SOCKET_PATH') or default_moulti_socket_path(instance)

def moulti_bind_path(abstract: bool) -> str:
	username = current_username()
	pid = os.getpid()
	socket_filename = f'moulti-client-{username}-{pid}.socket'
	return make_socket_path(socket_filename, abstract, printable=False)

def is_abstract_socket(socket_path: str) -> bool:
	return bool(socket_path and socket_path[0] == '\0')

def from_printable(socket_path: str) -> str:
	if socket_path and socket_path[0] == '@':
		socket_path = '\0' + socket_path[1:]
	return socket_path

def to_printable(socket_path: str) -> str:
	if is_abstract_socket(socket_path):
		socket_path = '@' + socket_path[1:]
	return socket_path

PRINTABLE_MOULTI_SOCKET = moulti_socket_path()
MOULTI_SOCKET = from_printable(PRINTABLE_MOULTI_SOCKET)

class MoultiProtocolException(Exception):
	pass

class MoultiConnectionClosedException(Exception):
	def __init__(self, expected_bytes: int = 0, anomaly: bool = True):
		self.expected_bytes = expected_bytes
		self.anomaly = anomaly
	def __str__(self) -> str:
		if self.anomaly:
			return f'connection closed by peer while expecting {self.expected_bytes} bytes'
		return 'connection closed by peer'

def moulti_unix_socket() -> Socket:
	return Socket(AF_UNIX, SOCK_STREAM, 0)

def moulti_listen(bind: str = MOULTI_SOCKET, backlog: int = 100, blocking: bool = False) -> tuple[Socket, bool]:
	try:
		server_socket = moulti_unix_socket()
		abstract = is_abstract_socket(bind)
		if not abstract:
			os.makedirs(os.path.dirname(bind), exist_ok=True)
		server_socket.bind(bind)
		server_socket.listen(backlog)
		server_socket.setblocking(blocking)
		return server_socket, abstract
	except Exception as exc:
		err = f'cannot listen on {to_printable(bind)} (with backlog={backlog} and blocking={blocking}): {exc}'
		raise MoultiProtocolException(err) from exc

def clean_socket(socket_path: str = PRINTABLE_MOULTI_SOCKET) -> None:
	if not socket_path.startswith('@'):
		try:
			os.unlink(socket_path)
		except FileNotFoundError:
			pass

def moulti_connect(address: str = MOULTI_SOCKET, bind: str | None = None) -> Socket:
	client_socket = moulti_unix_socket()
	abstract = is_abstract_socket(address)
	final_bind = bind or moulti_bind_path(abstract)
	client_socket.bind(final_bind)
	if not is_abstract_socket(final_bind):
		# As a pure client, we do not actually need to keep our socket file:
		clean_socket(final_bind)
	client_socket.connect(address)
	return client_socket

def getraddr(socket: Socket) -> str:
	"""
	Return the remote address of the given socket.
	"""
	try:
		raddr = socket.getpeername()
	except OSError as ose:
		raddr = f'unknown/errno={ose.errno}'
	if not isinstance(raddr, str):
		raddr = raddr.decode('utf-8')
	raddr = raddr.replace('\0', '@')
	return raddr + ':fd' + str(socket.fileno())

# This is a simple text-based TLV (type-length-value) protocol.
# preamble examples: ":JSON:0000000002048:" ":TXT_:0000000002048:"
PREAMBLE_FIXED_LENGTH = 20
PREAMBLE_REGEX = r'^:[A-Z_]{4}:[0-9]{13}:$'

def parse_preamble(preamble_data: bytes) -> tuple[str, int]:
	"""
	Parse a Moulti TLV preamble; return type and length or raise
	MoultiProtocolException.
	"""
	try:
		preamble = preamble_data.decode('ascii')
	except UnicodeDecodeError as ude:
		raise MoultiProtocolException('non-ASCII preamble: {str(preamble)}') from ude
	if not re.match(PREAMBLE_REGEX, preamble):
		raise MoultiProtocolException(f'invalid preamble: {preamble}')
	data_type = preamble[1:5].rstrip('_')
	if not data_type:
		raise MoultiProtocolException(f'invalid data type: {preamble[1:5]}')
	data_length = int(preamble[6:19])
	return data_type, data_length

def read_fixed_amount_from_socket(socket: Socket, amount: int) -> bytes:
	data = bytes()
	while amount > 0:
		new_data = socket.recv(amount)
		if not new_data:
			raise MoultiConnectionClosedException(amount)
		data += new_data
		amount -= len(new_data)
	return data

def write_fixed_amount_to_socket(socket: Socket, data: bytes) -> None:
	if not data:
		return
	to_send = len(data)
	sent = 0
	while to_send > 0:
		just_sent = socket.send(data[sent:])
		to_send -= just_sent
		sent += just_sent

def read_tlv_data_from_socket(socket: Socket, max_fds: int = 1) -> tuple[str, bytes, FDs]:
	file_descriptors: FDs = []
	if max_fds > 0:
		_, file_descriptors, _, _ = recv_fds(socket, 0, max_fds)

	try:
		preamble = read_fixed_amount_from_socket(socket, PREAMBLE_FIXED_LENGTH)
	except MoultiConnectionClosedException as mcce:
		mcce.anomaly = mcce.expected_bytes != PREAMBLE_FIXED_LENGTH
		raise mcce
	data_type, data_length = parse_preamble(preamble)
	return data_type, read_fixed_amount_from_socket(socket, data_length), file_descriptors

def assemble_tlv(data: bytes, data_type: str = 'JSON') -> bytes:
	data_type = data_type + '____'
	data_length = len(data)
	preamble = f':{data_type[0:4]}:{data_length:013}:'.encode('ascii')
	return preamble + data

def write_tlv_data_to_socket(socket: Socket, data: bytes, data_type: str = 'JSON', fds: FDs|None = None) -> None:
	to_send = assemble_tlv(data, data_type)
	if fds:
		# File descriptors are sent as ancillary data along with a regular non-empty message; operating systems may
		# require such messages be sent atomically; if buffers do not allow that, the system returns EMSGSIZE.
		# First, try to send the whole message:
		try:
			send_fds(socket, [to_send], fds)
		except OSError as outer_ose:
			if outer_ose.errno != errno.EMSGSIZE:
				raise
			# Got EMSGSIZE: try to send only the first byte, in a loop:
			first_byte = [to_send[0:1]]
			fds_sent = False
			while not fds_sent:
				try:
					send_fds(socket, first_byte, fds)
					fds_sent = True
				except OSError as inner_ose:
					if inner_ose.errno != errno.EMSGSIZE:
						raise
			# Send the rest of the message:
			write_fixed_amount_to_socket(socket, to_send[1:])
	else:
		write_fixed_amount_to_socket(socket, to_send)

def data_to_message(data: bytes) -> Message:
	try:
		string = data.decode('utf-8')
		message = json.loads(string)
		return message
	except UnicodeDecodeError as ude:
		raise MoultiProtocolException(f'received non UTF-8 {len(data)}-byte line') from ude
	except JSONDecodeError as jde:
		raise MoultiProtocolException('received non-JSON message') from jde

def message_to_data(message: Message) -> bytes:
	string = json.dumps(message)
	data = string.encode('utf-8')
	return data

def send_json_message(socket: Socket, message: Message, fds: FDs | None = None) -> str:
	# Setting a message id yields smaller replies as the server does not have to send the original data again:
	if 'msgid' not in message:
		message = {'msgid': str(uuid.uuid4()), **message}
	data = message_to_data(message)
	write_tlv_data_to_socket(socket, data, 'JSON', fds)
	return message['msgid']

def recv_json_message(socket: Socket, max_fds: int = 1) -> tuple[Message, FDs]:
	data_type, data, file_descriptors = read_tlv_data_from_socket(socket, max_fds)
	if data_type != 'JSON':
		raise MoultiProtocolException('received non-JSON message')
	return data_to_message(data), file_descriptors

def send_to_moulti(message: Message, wait_for_reply: bool = True) -> Message | None:
	with moulti_connect() as moulti_socket:
		msgid = send_json_message(moulti_socket, message)
		if wait_for_reply:
			reply, _ = recv_json_message(moulti_socket, 0)
			if msgid in reply and reply['msgid'] != msgid:
				raise MoultiProtocolException('expected message id {msgid} but received {reply["msgid"]}')
			return reply
	return None

class MoultiTLVReader:
	"""
	Helper that reads TLV data from a given socket each time read() is called.
	Each time a complete TLV dataset has been read, call a given callback.
	"""
	# pylint: disable=attribute-defined-outside-init
	def __init__(
		self,
		socket: Socket,
		raddr: str,
		callback: TLVCallback,
		log_callback: LogCallback|None = None,
		max_fds: int = 1
	):
		self.socket = socket
		self.raddr = raddr
		self.callback = callback
		self.log_callback = log_callback
		self.max_fds = max_fds
		self.reset()

	def reset(self) -> None:
		self.recv_fds_done = False
		self.file_descriptors: FDs = []

		self.recv_preamble_done = False
		self.preamble_data = bytes()
		self.data_type = ''
		self.data_length = 0
		self.data_value = bytes()

	def log(self, line: str) -> None:
		if self.log_callback:
			self.log_callback(line)

	def got_complete_tlv(self) -> None:
		data_type, data, file_descriptors = self.data_type, self.data_value, self.file_descriptors
		self.reset()
		self.callback(self.socket, self.raddr, data_type, data, file_descriptors)

	def read(self) -> None:
		try:
			# Read file descriptors:
			if self.max_fds > 0 and not self.recv_fds_done:
				_, self.file_descriptors, _, _ = recv_fds(self.socket, 0, self.max_fds)
				self.recv_fds_done = True
			# Read preamble:
			if not self.recv_preamble_done:
				amount = PREAMBLE_FIXED_LENGTH - len(self.preamble_data)
				while amount > 0:
					new_data = self.socket.recv(amount)
					if not new_data:
						# If we notice the client closed its connection here, it may be normal:
						anomaly = amount != PREAMBLE_FIXED_LENGTH
						raise MoultiConnectionClosedException(amount, anomaly=anomaly)
					self.preamble_data += new_data
					amount -= len(new_data)
				self.data_type, self.data_length = parse_preamble(self.preamble_data)
				self.recv_preamble_done = True
			# Read data:
			while self.data_length > 0:
				new_data = self.socket.recv(self.data_length)
				if not new_data:
					raise MoultiConnectionClosedException(self.data_length)
				self.data_value += new_data
				self.data_length -= len(new_data)
			self.got_complete_tlv()
		except BlockingIOError:
			# It turns out there is nothing (left) to read; better luck next read().
			pass

class MoultiTLVWriter:
	"""
	Helper that writes TLV data to a given socket each time write() is called.
	Each time a complete TLV dataset has been written, call a given callback.
	"""
	# pylint: disable=attribute-defined-outside-init
	def __init__(
		self,
		socket: Socket,
		raddr: str,
		callback: TLVCallback|None,
		log_callback: LogCallback|None = None,
	):
		self.socket = socket
		self.raddr = raddr
		self.callback = callback
		self.log_callback = log_callback
		self.buffers: deque = deque()
		self.reset()

	def reset(self) -> None:
		self.message_type: str|None = None
		self.message_data: bytes|None = None
		self.data: bytes|None = None
		self.written = 0

	def log(self, line: str) -> None:
		if self.log_callback:
			self.log_callback(line)

	def finished_writing(self) -> None:
		if self.callback is not None:
			assert self.message_type is not None # for mypy
			assert self.message_data is not None # for mypy
			self.callback(self.socket, self.raddr, self.message_type, self.message_data, [])
		self.reset()

	def write_message(self, data: bytes, data_type: str = 'JSON', immediate: bool = True) -> bool:
		"""
		Append the given message to the list of messages to be sent.
		If immediate is True (the default), call the write() method.
		Return True if all buffered messages were written, False if some messages or data remain to be written.
		"""
		self.buffers.appendleft((data, data_type))
		return self.write() if immediate else False

	def next(self) -> bool:
		if self.buffers:
			self.message = self.buffers.pop()
			self.data = assemble_tlv(*self.message)
			self.written = 0
			return True
		return False

	def write(self) -> bool:
		"""
		Write data and buffered messages until the network socket would block.
		Return True if all buffered messages were written, False if some messages or data remain to be written.
		"""
		try:
			while self.data is not None or self.next():
				assert self.data is not None # for mypy
				self.written += self.socket.send(self.data[self.written:])
				if self.written == len(self.data):
					self.finished_writing()
			# Everything was written:
			return True
		except BlockingIOError:
			# Cannot write without blocking; better luck next write().
			return False



from .common import moulti_test, wait_script_start, wait_script_end
assert moulti_test

def test_buttonquestion(moulti_test):
	command = ['buttonquestion.bash']
	async def run_before(pilot):
		await wait_script_start(pilot)
		await pilot.pause(.5)
		await pilot.press('ctrl+y') # next unanswered question
		await pilot.pause(.1)
		await pilot.press('tab', 'tab', 'enter') # hit 3rd button
		await wait_script_end(pilot)
	assert moulti_test(command=command, run_before=run_before)



from textual.containers import VerticalScroll

class VertScroll(VerticalScroll):
	"""
	Variant of Textual's VerticalScroll container that bears the
	"vertical_scrollbar_visible" class when its vertical scrollbar is visible.
	"""
	def watch_show_vertical_scrollbar(self) -> None:
		self.set_class(self.show_vertical_scrollbar, 'vertical_scrollbar_visible')



"""
Provide everything to reject unauthorized Moulti clients.
"""
from os import environ, getuid
from struct import calcsize, unpack
import sys
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from .app import Moulti
	from socket import socket as Socket

def ids_from_env(env_var_name: str) -> list[int]:
	try:
		return [int(xid) for xid in environ.get(env_var_name, '').split(',')]
	except Exception:
		return []

def get_unix_credentials(socket: 'Socket') -> tuple[int, int, int]:
	if sys.platform == 'linux':
		# struct ucred is { pid_t, uid_t, gid_t }
		struct_ucred = '3i'
		from socket import SOL_SOCKET, SO_PEERCRED # pylint: disable=import-outside-toplevel,no-name-in-module
		unix_credentials = socket.getsockopt(SOL_SOCKET, SO_PEERCRED, calcsize(struct_ucred))
		pid, uid, gid = unpack(struct_ucred, unix_credentials)
		return pid, uid, gid
	return -1, -1, -1

class MoultiSecurityPolicy:
	def __init__(self, app: 'Moulti'):
		self.app = app

		# By default, allow only the current uid to connect to Moulti:
		self.allowed_uids = [getuid()]
		self.allowed_gids = []

		# Make this behaviour configurable through environment variables:
		self.allowed_uids.extend(ids_from_env('MOULTI_ALLOWED_UID'))
		self.allowed_gids.extend(ids_from_env('MOULTI_ALLOWED_GID'))

	def check(self, socket: 'Socket') -> str:
		# Check Unix credentials (uid/gid) if and only if Moulti listens on an abstract socket.
		# Regular sockets are protected by file permissions.
		if self.app.server is not None and self.app.server.server_socket_is_abstract:
			_, uid, gid = get_unix_credentials(socket)
			allowed = uid in self.allowed_uids or gid in self.allowed_gids
			if not allowed:
				return f'invalid Unix credentials {uid}:{gid}'
		return ''



from moulti.widgets.step.tui import Step

def test_bytes_to_lines():
	# No data at all, no lines:
	assert Step.bytes_to_lines(b'') == ([], -1, b'', b'')

	# Empty line:
	assert Step.bytes_to_lines(b'\n') == ([''], 0, b'', b'')
	# One regular line:
	assert Step.bytes_to_lines(b'data\n') == (['data'], 4, b'', b'')
	# Multiple regular lines:
	assert Step.bytes_to_lines(b'data1\ndata02\n') == (['data1', 'data02'], 6, b'', b'')
	# Alternating regular and empty lines:
	assert Step.bytes_to_lines(b'data1\n\ndata02\n\ndata3\n') == (['data1', '', 'data02', '', 'data3'], 6, b'', b'')

	# Leftover only, no lines:
	assert Step.bytes_to_lines(b'leftover') == ([], -1, b'leftover', b'')
	# Empty line with leftover:
	assert Step.bytes_to_lines(b'\nleftover') == ([''], 0, b'leftover', b'')
	# One line with leftover:
	assert Step.bytes_to_lines(b'data\nleftover') == (['data'], 4, b'leftover', b'')
	# Multiple lines with leftover:
	assert Step.bytes_to_lines(b'data1\ndata02\nleftover') == (['data1', 'data02'], 6, b'leftover', b'')
	# Alternating regular and empty lines with leftover:
	assert Step.bytes_to_lines(b'data1\n\ndata02\n\ndata03\nleftover') == (['data1', '', 'data02', '', 'data03'], 6, b'leftover', b'')
